#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import unittest

from commoncode import urn


class URNTestCase(unittest.TestCase):

    def test_encode_license(self):
        u1 = urn.encode('license', key='somekey')
        assert u1 == 'urn:dje:license:somekey'

    def test_encode_owner(self):
        u1 = urn.encode('owner', name='somekey')
        assert u1 == 'urn:dje:owner:somekey'

    def test_encode_component(self):
        u1 = urn.encode('component', name='name', version='version')
        assert u1 == 'urn:dje:component:name:version'

    def test_encode_component_no_version(self):
        u1 = urn.encode('component', name='name', version='')
        assert u1 == 'urn:dje:component:name:'

    def test_encode_license_with_extra_fields_are_ignored(self):
        u1 = urn.encode('license', key='somekey', junk='somejunk')
        assert u1 == 'urn:dje:license:somekey'

    def test_encode_missing_field_raise_keyerror(self):
        with self.assertRaises(KeyError):
            urn.encode('license')

    def test_encode_missing_field_component_raise_keyerror(self):
        with self.assertRaises(KeyError):
            urn.encode('component', name='this')

    def test_encode_unknown_object_type_raise_keyerror(self):
        with self.assertRaises(KeyError):
            urn.encode('some', key='somekey')

    def test_encode_component_with_spaces_are_properly_quoted(self):
        u1 = urn.encode('component', name='name space',
                        version='version space')
        assert u1 == 'urn:dje:component:name+space:version+space'

    def test_encode_leading_and_trailing_spaces_are_trimmed_and_ignored(self):
        u1 = urn.encode(' component ', name=' name space    ',
                        version='''  version space ''')
        assert u1 == 'urn:dje:component:name+space:version+space'

    def test_encode_component_with_semicolon_are_properly_quoted(self):
        u1 = urn.encode('component', name='name:', version=':version')
        assert u1 == 'urn:dje:component:name%3A:%3Aversion'

    def test_encode_component_with_plus_are_properly_quoted(self):
        u1 = urn.encode('component', name='name+', version='version+')
        assert u1 == 'urn:dje:component:name%2B:version%2B'

    def test_encode_component_with_percent_are_properly_quoted(self):
        u1 = urn.encode('component', name='name%', version='version%')
        assert u1 == 'urn:dje:component:name%25:version%25'

    def test_encode_object_type_case_is_not_significant(self):
        u1 = urn.encode('license', key='key')
        u2 = urn.encode('lICENSe', key='key')
        assert u2 == u1

    def test_decode_component(self):
        u = 'urn:dje:component:name:version'
        parsed = ('component', {'name': 'name', 'version': 'version'})
        assert urn.decode(u) == parsed

    def test_decode_license(self):
        u = 'urn:dje:license:lic'
        parsed = ('license', {'key': 'lic'})
        assert urn.decode(u) == parsed

    def test_decode_org(self):
        u = 'urn:dje:owner:name'
        parsed = ('owner', {'name': 'name'})
        assert urn.decode(u) == parsed

    def test_decode_build_is_idempotent(self):
        u1 = urn.encode('component', owner__name='org%', name='name%',
                       version='version%')
        m, f = urn.decode(u1)
        u3 = urn.encode(m, **f)
        assert u3 == u1

    def test_decode_raise_exception_if_incorrect_prefix(self):
        with self.assertRaises(urn.URNValidationError):
            urn.decode('arn:dje:a:a')

    def test_decode_raise_exception_if_incorrect_ns(self):
        with self.assertRaises(urn.URNValidationError):
                urn.decode('urn:x:x:x')

    def test_decode_raise_exception_if_incorrect_prefix_or_ns(self):
        with self.assertRaises(urn.URNValidationError):
                urn.decode('x:x:x:x')

    def test_decode_raise_exception_if_too_short_license(self):
        with self.assertRaises(urn.URNValidationError):
            urn.decode('urn:dje:license')

    def test_decode_raise_exception_if_too_short_component(self):
        with self.assertRaises(urn.URNValidationError):
            urn.decode('urn:dje:component')

    def test_decode_raise_exception_if_too_long(self):
        with self.assertRaises(urn.URNValidationError):
            urn.decode('urn:dje:owner:o:n')

    def test_decode_raise_exception_if_too_long1(self):
        with self.assertRaises(urn.URNValidationError):
            urn.decode('urn:dje:component:o:n:v:junk')

    def test_decode_raise_exception_if_too_long2(self):
        with self.assertRaises(urn.URNValidationError):
            urn.decode('urn:dje:owner:org:junk')

    def test_decode_raise_exception_if_too_long3(self):
        with self.assertRaises(urn.URNValidationError):
            urn.decode('urn:dje:license:key:junk')

    def test_decode_raise_exception_if_unknown_object_type(self):
        with self.assertRaises(urn.URNValidationError):
            urn.decode('urn:dje:marshmallows:dsds')

    def test_decode_raise_exception_if_missing_object_type(self):
        with self.assertRaises(urn.URNValidationError):
            urn.decode('urn:dje::dsds')

    def test_encode_decode_is_idempotent(self):
        object_type = 'component'
        fields = {'name': 'SIP Servlets (MSS)', 'version': 'v 1.4.0.FINAL'}
        encoded = 'urn:dje:component:SIP+Servlets+%28MSS%29:v+1.4.0.FINAL'
        assert urn.encode(object_type, **fields) == encoded
        assert urn.decode(encoded) == (object_type, fields)



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#


def set_re_max_cache(max_cache=1000000):
    """
    Set re and fnmatch _MAXCACHE to 1Million items to cache compiled regex
    aggressively. Their default is a maximum of 100 items and many utilities and
    libraries use a lot of regexes: therefore 100 is not enough to benefit from
    caching.
    """
    import re
    import fnmatch

    remax = getattr(re, '_MAXCACHE', 0)
    if remax < max_cache:
        setattr(re, '_MAXCACHE', max_cache)

    fnmatchmax = getattr(fnmatch, '_MAXCACHE', 0)
    if fnmatchmax < max_cache:
        setattr(fnmatch, '_MAXCACHE', max_cache)


set_re_max_cache()



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import ctypes
import contextlib
import io
import os
from os import path

import logging
import signal
import subprocess

from commoncode.fileutils import get_temp_dir
from commoncode.system import on_posix
from commoncode.system import on_windows
from commoncode import text

"""
Wrapper for executing external commands in sub-processes which works
consistently the same way on POSIX and Windows OSes and can cope with the
capture of large command standard outputs without exhausting memory.
"""

logger = logging.getLogger(__name__)

TRACE = False

if TRACE:
    import sys
    logging.basicConfig(stream=sys.stdout)
    logger.setLevel(logging.DEBUG)

# current directory is the root dir of this library
curr_dir = path.dirname(path.dirname(path.abspath(__file__)))

PATH_ENV_VAR = 'PATH'
LD_LIBRARY_PATH = 'LD_LIBRARY_PATH'
DYLD_LIBRARY_PATH = 'DYLD_LIBRARY_PATH'


def execute(cmd_loc, args, cwd=None, env=None, to_files=False, log=TRACE):
    """
    Run a `cmd_loc` command with the `args` arguments list and return the return
    code, the stdout and stderr.

    To avoid RAM exhaustion, always write stdout and stderr streams to files.

    If `to_files` is False, return the content of stderr and stdout as ASCII
    strings. Otherwise, return the locations to the stderr and stdout temporary
    files.

    Run the command using the `cwd` current working directory with an `env` dict
    of environment variables.
    """
    assert cmd_loc
    full_cmd = [cmd_loc] + (args or [])

    # any shared object should be either in the PATH, the rpath or
    # side-by-side with the exceutable
    cmd_dir = os.path.dirname(cmd_loc)
    env = get_env(env, lib_dir=cmd_dir) or None
    cwd = cwd or curr_dir

    # temp files for stderr and stdout
    tmp_dir = get_temp_dir(prefix='cmd-')

    sop = path.join(tmp_dir, 'stdout')
    sep = path.join(tmp_dir, 'stderr')

    # shell==True is DANGEROUS but we are not running arbitrary commands
    # though we can execute commands that just happen to be in the path
    # See why we need it on Windows https://bugs.python.org/issue8557
    shell = True if on_windows else False

    if log:
        printer = logger.debug if TRACE else lambda x: print(x)
        printer(
            'Executing command %(cmd_loc)r as:\n%(full_cmd)r\nwith: env=%(env)r\n'
            'shell=%(shell)r\ncwd=%(cwd)r\nstdout=%(sop)r\nstderr=%(sep)r'
            % locals())

    proc = None
    rc = 100

    try:
        with io.open(sop, 'wb') as stdout, io.open(sep, 'wb') as stderr, pushd(cmd_dir):
            proc = subprocess.Popen(
                full_cmd,
                cwd=cwd,
                env=env,
                stdout=stdout,
                stderr=stderr,
                shell=shell,
                # -1 defaults bufsize to system bufsize
                bufsize=-1,
                universal_newlines=True,
            )
            stdout, stderr = proc.communicate()
            rc = proc.returncode if proc else 0
    finally:
        close(proc)

    if not to_files:
        # return output as ASCII string loaded from the output files
        with open(sop, 'rb') as so:
            sor = so.read()
            sop = text.toascii(sor).strip()

        with open(sep, 'rb') as se:
            ser = se.read()
            sep = text.toascii(ser).strip()

    return rc, sop, sep


def execute2(
    cmd_loc,
    args,
    lib_dir=None,
    cwd=None,
    env=None,
    to_files=False,
    log=TRACE,
):
    """
    DEPRECATED: DO NOT USE. Use execute() instead
    Run a `cmd_loc` command with the `args` arguments list and return the return
    code, the stdout and stderr.

    To avoid RAM exhaustion, always write stdout and stderr streams to files.

    If `to_files` is False, return the content of stderr and stdout as ASCII
    strings. Otherwise, return the locations to the stderr and stdout temporary
    files.

    Run the command using the `cwd` current working directory with an `env` dict
    of environment variables.
    """
    import warnings

    warnings.warn(
        "commoncode.command.execute2 is deprecated. Use execute() instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    return execute(cmd_loc, args, cwd, env, to_files, log)


def get_env(base_vars=None, lib_dir=None):
    """
    Return a dictionary of environment variables for command execution with
    appropriate DY/LD_LIBRARY_PATH path variables. Use the optional `base_vars`
    environment variables dictionary as a base if provided. Note: if `base_vars`
    contains DY/LD_LIBRARY_PATH variables these will be overwritten. On POSIX,
    add `lib_dir` as DY/LD_LIBRARY_PATH-like path if provided.
    """
    env_vars = {}
    if base_vars:
        env_vars.update(base_vars)

    # Create and add LD environment variables
    if lib_dir and on_posix:
        new_path = f'{lib_dir}'
        # on Linux/posix
        ld_lib_path = os.environ.get(LD_LIBRARY_PATH)
        env_vars.update({LD_LIBRARY_PATH: update_path_var(ld_lib_path, new_path)})
        # on Mac, though LD_LIBRARY_PATH should work too
        dyld_lib_path = os.environ.get(DYLD_LIBRARY_PATH)
        env_vars.update({DYLD_LIBRARY_PATH: update_path_var(dyld_lib_path, new_path)})

    env_vars = {text.as_unicode(k): text.as_unicode(v) for k, v in env_vars.items()}

    return env_vars


def close(proc):
    """
    Close a `proc` process opened pipes and kill the process.
    """
    if not proc:
        return

    def close_pipe(p):
        if not p:
            return
        try:
            p.close()
        except IOError:
            pass

    close_pipe(getattr(proc, 'stdin', None))
    close_pipe(getattr(proc, 'stdout', None))
    close_pipe(getattr(proc, 'stderr', None))

    try:
        # Ensure process death otherwise proc.wait may hang in some cases
        # NB: this will run only on POSIX OSes supporting signals
        os.kill(proc.pid, signal.SIGKILL)  # NOQA
    except:
        pass

    # This may slow things down a tad on non-POSIX Oses but is safe:
    # this calls os.waitpid() to make sure the process is dead
    proc.wait()


def load_shared_library(dll_loc, *args):
    """
    Return the loaded shared library object from the ``dll_loc`` location.
    """
    if not dll_loc or not path.exists(dll_loc):
        raise ImportError(f'Shared library does not exists: dll_loc: {dll_loc}')

    if not isinstance(dll_loc, str):
        dll_loc = os.fsdecode(dll_loc)

    lib = None

    dll_dir = os.path.dirname(dll_loc)
    try:
        with pushd(dll_dir):
            lib = ctypes.CDLL(dll_loc)
    except OSError as e:
        from pprint import pformat
        import traceback
        msgs = tuple([
            f'ctypes.CDLL("{dll_loc}")',
            'os.environ:\n{}'.format(pformat(dict(os.environ))),
            traceback.format_exc(),
        ])
        raise Exception(msgs) from e

    if lib and lib._name:
        return lib

    raise Exception(f'Failed to load shared library with ctypes: {dll_loc}')


@contextlib.contextmanager
def pushd(path):
    """
    Context manager to change the current working directory to `path`.
    """
    original_cwd = os.getcwd()
    if not path:
        path = original_cwd
    try:
        os.chdir(path)
        yield os.getcwd()
    finally:
        os.chdir(original_cwd)


def update_path_var(existing_path_var, new_path):
    """
    Return an updated value for the `existing_path_var` PATH-like environment
    variable value  by adding `new_path` to the front of that variable if
    `new_path` is not already part of this PATH-like variable.
    """
    if not new_path:
        return existing_path_var

    existing_path_var = existing_path_var or ''

    existing_path_var = os.fsdecode(existing_path_var)
    new_path = os.fsdecode(new_path)

    path_elements = existing_path_var.split(os.pathsep)

    if not path_elements:
        updated_path_var = new_path

    elif new_path not in path_elements:
        # add new path to the front of the PATH env var
        path_elements.insert(0, new_path)
        updated_path_var = os.pathsep.join(path_elements)

    else:
        # new path is already in PATH, change nothing
        updated_path_var = existing_path_var

    if not isinstance(updated_path_var, str):
        updated_path_var = os.fsdecode(updated_path_var)

    return updated_path_var


PATH_VARS = DYLD_LIBRARY_PATH, LD_LIBRARY_PATH, 'PATH',


def searchable_paths(env_vars=PATH_VARS):
    """
    Return a list of directories where to search "in the PATH" in the provided
    ``env_vars`` list of PATH-like environment variables.
    """
    dirs = []
    for env_var in env_vars:
        value = os.environ.get(env_var, '') or ''
        dirs.extend(value.split(os.pathsep))
    dirs = [os.path.realpath(d.strip()) for d in dirs if d.strip()]
    return tuple(d for d in dirs if os.path.isdir(d))


def find_in_path(filename, searchable_paths=searchable_paths()):
    """
    Return the location of a ``filename`` found in the ``searchable_paths`` list
    of directory or None.
    """
    for path in searchable_paths:
        location = os.path.join(path, filename)
        if os.path.exists(location):
            return location



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from datetime import datetime
from datetime import tzinfo
from functools import update_wrapper
from functools import wraps
from time import time

"""
Time is of the essence: path safe time stamps creation and conversion to
datetime objects.
"""


class UTC(tzinfo):
    """UTC timezone"""

    def utcoffset(self, dt):  # NOQA
        return None

    def tzname(self, dt):  # NOQA
        return 'UTC'

    def dst(self, dt):  # NOQA
        return None


def time2tstamp(dt=None, path_safe=True):
    """
    Return a timestamp representing the datetime object (assumed to be in UTC
    time) or the current UTC time (if dt == None) formatted using the ISO 8601
    standard as a basis, extended to be path safe is path_safe is True.

    The Python isoformat returns a time stamp that complies with this standard
    but has limitations when used in a file or directory name. Here we
    transform the returned time stamp such that the result still complies with
    the ISO standard and can be safely used as part of a of file or directory
    name in a portable and OS safe fashion including on Windows where colons
    are not allowed in file names, or on posix where / denotes a path segment
    separator.

    For times, the ISO 8601 format specifies either a colon : (extended format)
    or nothing as a separator (basic format). Here Python defaults to using a
    colon. We therefore remove all the colons to be safe across filesystems. (a
    colon is not a valid path char on Windows)

    Another character may show up in the ISO representation such as / for time
    intervals. We could replace the forward slash with a double hyphen (--) as
    a separator instead (see Section 4.4.2 of the ISO standard). However since
    there are several places where hyphens are used, this makes it difficult to
    parse back. Instead we use an _ (underscore) to make the time stamp easier
    to convert back to a datetime object.
    """
    # TODO: check that the dt is effectively in UTC
    datim = dt or datetime.utcnow()
    iso = datim.isoformat()
    if path_safe:
        iso = iso.replace(':', '').replace('/', '_')
    return iso


def tstamp2time(stamp):
    """
    Convert a UTC timestamp to a datetime object.
    """
    # handle both basic and extended formats
    tformat = '%Y-%m-%dT%H%M%S' if stamp[4] == '-' else '%Y%m%dT%H%M%S'
    # normalize
    dt_ms = stamp.strip().replace('Z', '').replace(':', '')

    dt_ms = dt_ms.split('.')
    isodatim = dt_ms[0]
    datim = datetime.strptime(isodatim, tformat)
    # all stamps must be UTC
    datim = datim.replace(tzinfo=UTC())

    # deal with optional microsec
    try:
        microsec = dt_ms[1]
    except:
        microsec = None
    if microsec:
        microsec = int(microsec)
        if 0 <= microsec <= 999999:
            datim = datim.replace(microsecond=microsec)
    return datim


def timed(fun):
    """
    Decorate `fun` callable to return a tuple of (timing, result) where timing
    is a function execution time in seconds as a float and result is the value
    returned by calling `fun`.

    Note: this decorator will not work as expected for functions that return
    generators.
    """

    @wraps(fun)
    def _timed(*args, **kwargs):
        start = time()
        result = fun(*args, **kwargs)
        return time() - start, result

    return update_wrapper(_timed, fun)



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import calendar
from datetime import datetime
import os


def isoformat(utc_date):
    return datetime.isoformat(utc_date).replace('T', ' ')


def get_file_mtime(location, iso=True):
    """
    Return a string containing the last modified date of a file formatted
    as an ISO time stamp if ISO is True or as a raw number since epoch.
    """
    date = ''
    # FIXME: use file types
    if not os.path.isdir(location):
        mtime = os.stat(location).st_mtime
        if iso:
            utc_date = datetime.utcfromtimestamp(mtime)
            date = isoformat(utc_date)
        else:
            date = str(mtime)
    return date


def secs_from_epoch(d):
    """
    Return a number of seconds since epoch for a date time stamp
    """
    # FIXME: what does this do?
    return calendar.timegm(datetime.strptime(d.split('.')[0],
                                    '%Y-%m-%d %H:%M:%S').timetuple())



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import os

import click
click.disable_unicode_literals_warning = True
from click.testing import CliRunner

from commoncode.testcase import FileDrivenTesting
from commoncode.cliutils import fixed_width_file_name
from commoncode.cliutils import GroupedHelpCommand
from commoncode.cliutils import PluggableCommandLineOption


class TestFixedWidthFilename(FileDrivenTesting):

    def test_fixed_width_file_name_with_file_name_larger_than_max_length_is_shortened(self):
        test = fixed_width_file_name('0123456789012345678901234.c', 25)
        expected = '0123456789...5678901234.c'
        assert test == expected

    def test_fixed_width_file_name_with_file_name_smaller_than_max_length_is_not_shortened(self):
        file_name = '0123456789012345678901234.c'
        test = fixed_width_file_name(file_name, max_length=50)
        assert test == file_name

    def test_fixed_width_file_name_with_file_name_at_max_length_is_not_shortened(self):
        test = fixed_width_file_name('01234567890123456789012.c', 25)
        expected = '01234567890123456789012.c'
        assert test == expected

    def test_fixed_width_file_name_with_file_name_smaller_than_max_length_not_shortened(self):
        test = fixed_width_file_name('0123456789012345678901.c', 25)
        expected = '0123456789012345678901.c'
        assert test == expected

    def test_fixed_width_file_name_with_none_filename_return_empty_string(self):
        test = fixed_width_file_name(None, 25)
        expected = ''
        assert test == expected

    def test_fixed_width_file_name_without_extension(self):
        test = fixed_width_file_name('012345678901234567890123456', 25)
        expected = '01234567890...67890123456'
        assert test == expected

    def test_fixed_width_file_name_with_posix_path_without_shortening(self):
        test = fixed_width_file_name('C/Documents_and_Settings/Boki/Desktop/head/patches/drupal6/drupal.js', 25)
        expected = 'drupal.js'
        assert test == expected

    def test_fixed_width_file_name_with_posix_path_with_shortening(self):
        test = fixed_width_file_name('C/Documents_and_Settings/Boki/Desktop/head/patches/drupal6/012345678901234567890123.c', 25)
        expected = '0123456789...4567890123.c'
        assert test == expected

    def test_fixed_width_file_name_with_win_path_without_shortening(self):
        test = fixed_width_file_name('C\\:Documents_and_Settings\\Boki\\Desktop\\head\\patches\\drupal6\\drupal.js', 25)
        expected = 'drupal.js'
        assert test == expected

    def test_fixed_width_file_name_with_win_path_with_shortening(self):
        test = fixed_width_file_name('C\\:Documents_and_Settings\\Boki\\Desktop\\head\\patches\\drupal6\\012345678901234567890123.c', 25)
        expected = '0123456789...4567890123.c'
        assert test == expected

    def test_fixed_width_file_name_with_very_small_file_name_and_long_extension(self):
        test = fixed_width_file_name('abc.abcdef', 5)
        # FIXME: what is expected is TBD
        expected = ''
        assert test == expected


class TestGroupedHelpCommand(FileDrivenTesting):

    test_data_dir = os.path.join(os.path.dirname(__file__), 'data')

    def test_GroupedHelpCommand_help_group_and_sort_order_without_custom_class(self):

        @click.command(name='scan', cls=GroupedHelpCommand)
        @click.option('--opt', is_flag=True, help='Help text for option')
        def scan(opt):
            pass

        runner = CliRunner()
        result = runner.invoke(scan, ['--help'])
        from commoncode.cliutils import MISC_GROUP
        assert MISC_GROUP in result.output
        assert  '--opt   Help text for option' in result.output

    def test_GroupedHelpCommand_with_help_group_and_sort_order_with_custom_class(self):

        @click.command(name='scan', cls=GroupedHelpCommand)
        @click.option('--opt', is_flag=True, sort_order=10,
                      help='Help text for option', cls=PluggableCommandLineOption)
        def scan(opt):
            pass

        runner = CliRunner()
        result = runner.invoke(scan, ['--help'])
        from commoncode.cliutils import MISC_GROUP
        assert MISC_GROUP + ':\n    --opt   Help text for option\n' in result.output

    def test_GroupedHelpCommand_help_with_group(self):
        from commoncode.cliutils import CORE_GROUP

        @click.command(name='scan', cls=GroupedHelpCommand)
        @click.option(
            '--opt',
            is_flag=True,
            help='Help text for option',
            help_group=CORE_GROUP,
            cls=PluggableCommandLineOption,
        )
        def scan(opt):
            pass

        runner = CliRunner()
        result = runner.invoke(scan, ['--help'])
        assert CORE_GROUP + ':\n    --opt  Help text for option\n' in result.output



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import shutil
import sys

import click

from click.utils import echo
from click.termui import style
from click.types import BoolParamType
# FIXME: this is NOT API
from click._termui_impl import ProgressBar

from commoncode.fileutils import file_name
from commoncode.fileutils import splitext
from commoncode.text import toascii

# Tracing flags
TRACE = False


def logger_debug(*args):
    pass


if TRACE:
    import logging
    logger = logging.getLogger(__name__)
    logging.basicConfig(stream=sys.stdout)
    logger.setLevel(logging.DEBUG)

    def logger_debug(*args):
        return logger.debug(' '.join(isinstance(a, str)
                                     and a or repr(a) for a in args))

"""
Command line UI utilities for improved options, help and progress reporting.
"""


class BaseCommand(click.Command):
    """
    An enhanced click Command working around some Click quirk.
    """

    # override this in sub-classes with a command-specific message such as
    # "Try 'scancode --help' for help on options and arguments."
    short_usage_help = ''

    def get_usage(self, ctx):
        """
        Ensure that usage points to the --help option explicitly.
        Workaround click issue https://github.com/mitsuhiko/click/issues/393
        """
        return super(BaseCommand, self).get_usage(ctx) + self.short_usage_help

    def main(
        self, args=None, prog_name=None, complete_var=None,
        standalone_mode=True, **extra,
    ):
        """
        Workaround click 4.0 bug https://github.com/mitsuhiko/click/issues/365
        """
        return click.Command.main(
            self,
            args=args,
            prog_name=self.name,
            complete_var=complete_var,
            standalone_mode=standalone_mode,
            **extra,
        )


class GroupedHelpCommand(BaseCommand):
    """
    A command class that is aware of pluggable options that provides enhanced
    help where each option is grouped by group in the help.
    """

    short_usage_help = '''
Try the '--help' option for help on options and arguments.'''

    def __init__(
        self, name, context_settings=None, callback=None, params=None,
        help=None,  # NOQA
        epilog=None, short_help=None,
        options_metavar='[OPTIONS]', add_help_option=True,
        plugin_options=(),
    ):
        """
        Create a new GroupedHelpCommand using the `plugin_options` list of
        PluggableCommandLineOption instances.
        """

        super(GroupedHelpCommand, self).__init__(
            name,
            context_settings,
            callback,
            params,
            help,
            epilog,
            short_help,
            options_metavar,
            add_help_option,
        )

        # this makes the options "known" to the command
        self.params.extend(plugin_options)

    def format_options(self, ctx, formatter):
        """
        Overridden from click.Command to write all options into the formatter in
        help_groups they belong to. If a group is not specified, add the option
        to MISC_GROUP group.
        """
        # this mapping defines the CLI help presentation order
        help_groups = dict([
            (SCAN_GROUP, []),
            (OTHER_SCAN_GROUP, []),
            (SCAN_OPTIONS_GROUP, []),
            (OUTPUT_GROUP, []),
            (OUTPUT_FILTER_GROUP, []),
            (OUTPUT_CONTROL_GROUP, []),
            (PRE_SCAN_GROUP, []),
            (POST_SCAN_GROUP, []),
            (CORE_GROUP, []),
            (MISC_GROUP, []),
            (DOC_GROUP, []),
        ])

        for param in self.get_params(ctx):
            # Get the list of option's name and help text
            help_record = param.get_help_record(ctx)
            if not help_record:
                continue
            # organize options by group
            help_group = getattr(param, 'help_group', MISC_GROUP)
            sort_order = getattr(param, 'sort_order', 100)
            help_groups[help_group].append((sort_order, help_record))

        with formatter.section('Options'):
            for group, help_records in help_groups.items():
                if not help_records:
                    continue
                with formatter.section(group):
                    sorted_records = [hr for _, hr in sorted(help_records)]
                    formatter.write_dl(sorted_records)


# overriden and copied from Click to work around Click woes for
# https://github.com/nexB/scancode-toolkit/issues/2583
class DebuggedProgressBar(ProgressBar):

    # overriden and copied from Click to work around Click woes for
    # https://github.com/nexB/scancode-toolkit/issues/2583
    def make_step(self, n_steps):
        # always increment
        self.pos += n_steps or 1
        super(DebuggedProgressBar, self).make_step(n_steps)

    # overriden and copied from Click to work around Click woes for
    # https://github.com/nexB/scancode-toolkit/issues/2583
    def generator(self):
        if self.is_hidden:
            yield from self.iter
        else:
            for rv in self.iter:
                self.current_item = rv
                self.update(1)
                self.render_progress()
                yield rv

            self.finish()
            self.render_progress()


class EnhancedProgressBar(DebuggedProgressBar):
    """
    Enhanced progressbar ensuring that nothing is displayed when the bar is hidden.
    """

    def render_progress(self):
        if not self.is_hidden:
            return super(EnhancedProgressBar, self).render_progress()


class ProgressLogger(ProgressBar):
    """
    A subclass of Click ProgressBar providing a verbose line-by-line progress
    reporting.

    In contrast with the progressbar the label, percent, ETA, pos, bar_template
    and other formatting options are ignored.

    Progress information are printed as-is and no LF is added. The caller must
    provide an item_show_func to display some content and this must terminated
    with a line feed if needed.

    If no item_show_func is provided a simple dot is printed for each event.
    """

    def __init__(self, *args, **kwargs):
        super(ProgressLogger, self).__init__(*args, **kwargs)
        self.is_hidden = False

    def render_progress(self):
        line = self.format_progress_line()
        if line:
            # only add new lines if there is an item_show_func
            nl = bool(self.item_show_func)
            echo(line, file=self.file, nl=nl, color=self.color)
            self.file.flush()

    def format_progress_line(self):
        if self.item_show_func:
            item_info = self.item_show_func(self.current_item)
        else:
            item_info = '.'
        if item_info:
            return item_info

    def render_finish(self):
        self.file.flush()


BAR_WIDTH = 20
BAR_SEP_LEN = len(' ')


def progressmanager(
    iterable=None,
    length=None,
    fill_char='#',
    empty_char='-',
    bar_template=None,
    info_sep=' ',
    show_eta=False,
    show_percent=False,
    show_pos=True,
    item_show_func=None,
    label=None,
    file=None,
    color=None,  # NOQA
    update_min_steps=0,
    width=BAR_WIDTH,
    verbose=False,
):
    """
    Return an iterable context manager showing progress as a progress bar
    (default) or item-by-item log (if verbose is True) while iterating.

    It's arguments are similar to Click.termui.progressbar with these new
    arguments added at the end of the signature:

    :param verbose:  if True, display a progress log. Otherwise, a progress bar.
    """
    if verbose:
        progress_class = ProgressLogger
    else:
        progress_class = EnhancedProgressBar
        bar_template = (
            '[%(bar)s]' + ' ' + '%(info)s'
            if bar_template is None else bar_template
        )

    kwargs = dict(
        iterable=iterable,
        length=length,
        fill_char=fill_char,
        empty_char=empty_char,
        bar_template=bar_template,
        info_sep=info_sep,
        show_eta=show_eta,
        show_percent=show_percent,
        show_pos=show_pos,
        item_show_func=item_show_func,
        label=label,
        file=file,
        color=color,
        width=width,
    )

    # Check if we have an "update_min_steps" argument that was introduced in
    # Click 8. See https://github.com/pallets/click/pull/1698
    # Note that we use this argument on Click 8 in order to fix a regression
    # that this same PR introduced by Click and tracked originally at
    # https://github.com/nexB/scancode-toolkit/issues/2583
    # Here we create a dummy progress_class and then for the attribute presence.
    pb = progress_class([])
    if hasattr(pb, 'update_min_steps'):
        kwargs['update_min_steps'] = update_min_steps

    return progress_class(**kwargs)


def fixed_width_file_name(path, max_length=25):
    """
    Return a fixed width file name of at most `max_length` characters computed
    from the `path` string and usable for fixed width display. If the `path`
    file name is longer than `max_length`, the file name is truncated in the
    middle using three dots "..." as an ellipsis and the ext is kept.

    For example:
    >>> fwfn = fixed_width_file_name('0123456789012345678901234.c')
    >>> assert fwfn == '0123456789...5678901234.c'
    >>> fwfn = fixed_width_file_name('some/path/0123456789012345678901234.c')
    >>> assert fwfn == '0123456789...5678901234.c'
    >>> fwfn = fixed_width_file_name('some/sort.c')
    >>> assert fwfn == 'sort.c'
    >>> fwfn = fixed_width_file_name('some/123456', max_length=5)
    >>> assert fwfn == ''
    """
    if not path:
        return ''

    # get the path as unicode for display!
    filename = file_name(path)
    if len(filename) <= max_length:
        return filename
    base_name, ext = splitext(filename)
    dots = 3
    len_ext = len(ext)
    remaining_length = max_length - len_ext - dots

    if remaining_length < 5  or remaining_length < (len_ext + dots):
        return ''

    prefix_and_suffix_length = abs(remaining_length // 2)
    prefix = base_name[:prefix_and_suffix_length]
    ellipsis = dots * '.'
    suffix = base_name[-prefix_and_suffix_length:]
    return '{prefix}{ellipsis}{suffix}{ext}'.format(**locals())


def file_name_max_len(used_width=BAR_WIDTH + 1 + 7 + 1 + 8 + 1):
    """
    Return the max length of a path given the current terminal width.

    A progress bar is composed of these elements:
      [-----------------------------------#]  1667  Scanned: tu-berlin.yml
    - the bar proper which is BAR_WIDTH characters
    - one space
    - the number of files. We set it to 7 chars, eg. 9 999 999 files
    - one space
    - the word Scanned: 8 chars
    - one space
    - the file name proper
    The space usage is therefore:
        BAR_WIDTH + 1 + 7 + 1 + 8 + 1 + the file name length
    """
    term_width, _height = shutil.get_terminal_size()
    max_filename_length = term_width - used_width
    return max_filename_length


def path_progress_message(item, verbose=False, prefix='Scanned: '):
    """
    Return a styled message suitable for progress display when processing a path
    for an `item` tuple of (location, rid, scan_errors, *other items)
    """
    if not item:
        return ''
    location = item[0]
    errors = item[2]
    location = toascii(location)
    progress_line = location
    if not verbose:
        max_file_name_len = file_name_max_len()
        # do not display a file name in progress bar if there is no space available
        if max_file_name_len <= 10:
            return ''
        progress_line = fixed_width_file_name(location, max_file_name_len)

    color = 'red' if errors else 'green'
    return style(prefix) + style(progress_line, fg=color)


# CLI help groups
SCAN_GROUP = 'primary scans'
SCAN_OPTIONS_GROUP = 'scan options'
OTHER_SCAN_GROUP = 'other scans'
OUTPUT_GROUP = 'output formats'
OUTPUT_CONTROL_GROUP = 'output control'
OUTPUT_FILTER_GROUP = 'output filters'
PRE_SCAN_GROUP = 'pre-scan'
POST_SCAN_GROUP = 'post-scan'
MISC_GROUP = 'miscellaneous'
DOC_GROUP = 'documentation'
CORE_GROUP = 'core'


class PluggableCommandLineOption(click.Option):
    """
    An option with extra args and attributes to control CLI help options
    grouping, co-required and conflicting options (e.g. mutually exclusive).
    This option is also pluggable e.g. providable by a plugin.
    """

    # args are from Click 6.7
    def __init__(
        self,
        param_decls=None,
        show_default=False,
        prompt=False,
        confirmation_prompt=False,
        hide_input=False,
        is_flag=None,
        flag_value=None,
        multiple=False,
        count=False,
        allow_from_autoenv=True,
        type=None,  # NOQA
        help=None,  # NOQA
        hidden=False,
        # custom additions #
        # a string that set the CLI help group for this option
        help_group=MISC_GROUP,
        # a relative sort order number (integer or float) for this
        # option within a help group: the sort is by increasing
        # sort_order then by option declaration.
        sort_order=100,
        # a sequence of other option name strings that this option
        # requires to be set
        required_options=(),
        # a sequence of other option name strings that this option
        # conflicts with if they are set
        conflicting_options=(),
        **kwargs
    ):
        super(PluggableCommandLineOption, self).__init__(
            param_decls=param_decls,
            show_default=show_default,
            prompt=prompt,
            confirmation_prompt=confirmation_prompt,
            hide_input=hide_input,
            is_flag=is_flag,
            flag_value=flag_value,
            multiple=multiple,
            count=count,
            allow_from_autoenv=allow_from_autoenv,
            type=type,
            help=help,
            **kwargs
        )

        self.help_group = help_group
        self.sort_order = sort_order
        self.required_options = required_options
        self.conflicting_options = conflicting_options
        self.hidden = hidden

    def __repr__(self, *args, **kwargs):
        name = self.name
        opt = self.opts[-1]
        help_group = self.help_group
        required_options = self.required_options
        conflicting_options = self.conflicting_options

        return (
            'PluggableCommandLineOption<name=%(name)r, '
            'required_options=%(required_options)r, '
            'conflicting_options=%(conflicting_options)r>' % locals()
        )

    def validate_dependencies(self, ctx, value):
        """
        Validate `value` against declared `required_options` or
        `conflicting_options` dependencies.
        """
        _validate_option_dependencies(ctx, self, value, self.required_options, required=True)
        _validate_option_dependencies(ctx, self, value, self.conflicting_options, required=False)

    def get_help_record(self, ctx):
        if not self.hidden:
            return click.Option.get_help_record(self, ctx)


def validate_option_dependencies(ctx):
    """
    Validate all PluggableCommandLineOption dependencies in the `ctx` Click
    context. Ignore eager flags.
    """
    values = ctx.params
    if TRACE:
        logger_debug('validate_option_dependencies: values:')
        for va in sorted(values.items()):
            logger_debug('  ', va)

    for param in ctx.command.params:
        if param.is_eager:
            continue
        if not isinstance(param, PluggableCommandLineOption):
            if TRACE:
                logger_debug('  validate_option_dependencies: skip param:', param)
            continue
        value = values.get(param.name)
        if TRACE:
            logger_debug('  validate_option_dependencies: param:', param, 'value:', value)
        param.validate_dependencies(ctx, value)


def _validate_option_dependencies(ctx, param, value, other_option_names, required=False):
    """
    Validate the `other_option_names` option dependencies and return a
    UsageError if the `param` `value` is set to a not-None non-default value and
    if:
    - `required` is True and the `other_option_names` options are not set with a
       not-None value in the `ctx` context.
    - `required` is False and any of the `other_option_names` options are set
       with a not-None, non-default value in the `ctx` context.
    """
    if not other_option_names:
        return

    def _is_set(_value, _param):
        if _param.type in (bool, BoolParamType):
            return _value

        if _param.multiple:
            empty = (_value and len(_value) == 0) or not _value
        else:
            empty = _value is None

        return bool(not empty and _value != _param.default)

    is_set = _is_set(value, param)

    if TRACE:
        logger_debug()
        logger_debug('Checking param:', param)
        logger_debug('  value:', value, 'is_set:' , is_set)

    if not is_set:
        return

    oparams_by_name = {oparam.name: oparam for oparam in ctx.command.params}
    oparams = []
    missing_onames = []

    for oname in other_option_names:
        oparam = oparams_by_name.get(oname)
        if not oparam:
            missing_onames.append(oparam)
        else:
            oparams.append(oparam)

    if TRACE:
        logger_debug()
        logger_debug('  Available other params:')
        for oparam in oparams:
            logger_debug('    other param:', oparam)
            logger_debug('      value:', ctx.params.get(oparam.name))
        if required:
            logger_debug('    missing names:', missing_onames)

    if required and missing_onames:
        opt = param.opts[-1]
        oopts = [oparam.opts[-1] for oparam in oparams]
        omopts = ['--' + oname.replace('_', '-') for oname in missing_onames]
        oopts.extend(omopts)
        oopts = ', '.join(oopts)
        msg = ('The option %(opt)s requires the option(s) %(all_opts)s.'
               'and is missing %(omopts)s. '
               'You must set all of these options if you use this option.' % locals())
        raise click.UsageError(msg)

    if TRACE:
        logger_debug()
        logger_debug('  Checking other params:')

    opt = param.opts[-1]

    for oparam in oparams:
        ovalue = ctx.params.get(oparam.name)
        ois_set = _is_set(ovalue, oparam)

        if TRACE:
            logger_debug('    Checking oparam:', oparam)
            logger_debug('      value:', ovalue, 'ois_set:' , ois_set)

        # by convention the last opt is the long form
        oopt = oparam.opts[-1]
        oopts = ', '.join(oparam.opts[-1] for oparam in oparams)
        all_opts = '%(opt)s and %(oopts)s' % locals()
        if required and not ois_set:
            msg = ('The option %(opt)s requires the option(s) %(oopts)s '
                   'and is missing %(oopt)s. '
                   'You must set all of these options if you use this option.' % locals())
            raise click.UsageError(msg)

        if not required  and ois_set:
            msg = ('The option %(opt)s cannot be used together with the %(oopts)s option(s) '
                   'and %(oopt)s is used. '
                   'You can set only one of these options at a time.' % locals())
            raise click.UsageError(msg)



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from functools import partial
import os
from os import path
import gzip
import tarfile
import zipfile

from commoncode.system import on_windows

"""
Mimimal tar and zip file handling, primarily for testing.
"""


def _extract_tar_raw(test_path, target_dir, to_bytes, *args, **kwargs):
    """
    Raw simplified extract for certain really weird paths and file
    names.
    """
    tar = None
    try:
        tar = tarfile.open(test_path)
        tar.extractall(path=target_dir)
    finally:
        if tar:
            tar.close()


extract_tar_raw = partial(_extract_tar_raw, to_bytes=True)

extract_tar_uni = partial(_extract_tar_raw, to_bytes=False)


def extract_tar(location, target_dir, verbatim=False, *args, **kwargs):
    """
    Extract a tar archive at location in the target_dir directory.
    If `verbatim` is True preserve the permissions.
    """
    # always for using bytes for paths on all OSses... tar seems to use bytes internally
    # and get confused otherwise
    location = os.fsencode(location)
    with open(location, 'rb') as input_tar:
        tar = None
        try:
            tar = tarfile.open(fileobj=input_tar)
            tarinfos = tar.getmembers()
            to_extract = []
            for tarinfo in tarinfos:
                if tar_can_extract(tarinfo, verbatim):
                    if not verbatim:
                        tarinfo.mode = 0o755
                    to_extract.append(tarinfo)
            tar.extractall(target_dir, members=to_extract)
        finally:
            if tar:
                tar.close()


def extract_zip(location, target_dir, *args, **kwargs):
    """
    Extract a zip archive file at location in the target_dir directory.
    """
    if not path.isfile(location) and zipfile.is_zipfile(location):
        raise Exception('Incorrect zip file %(location)r' % locals())

    with zipfile.ZipFile(location) as zipf:
        for info in zipf.infolist():
            name = info.filename
            content = zipf.read(name)
            target = path.join(target_dir, name)
            if not path.exists(path.dirname(target)):
                os.makedirs(path.dirname(target))
            if not content and target.endswith(path.sep):
                if not path.exists(target):
                    os.makedirs(target)
            if not path.exists(target):
                with open(target, 'wb') as f:
                    f.write(content)


def extract_zip_raw(location, target_dir, *args, **kwargs):
    """
    Extract a zip archive file at location in the target_dir directory.
    Use the builtin extractall function
    """
    if not path.isfile(location) and zipfile.is_zipfile(location):
        raise Exception('Incorrect zip file %(location)r' % locals())

    with zipfile.ZipFile(location) as zipf:
        zipf.extractall(path=target_dir)


def tar_can_extract(tarinfo, verbatim):
    """
    Return True if a tar member can be extracted to handle OS specifics.
    If verbatim is True, always return True.
    """
    if tarinfo.ischr():
        # never extract char devices
        return False

    if verbatim:
        # extract all on all OSse
        return True

    # FIXME: not sure hard links are working OK on Windows
    include = tarinfo.type in tarfile.SUPPORTED_TYPES
    exclude = tarinfo.isdev() or (on_windows and tarinfo.issym())

    if include and not exclude:
        return True


def get_gz_compressed_file_content(location):
    """
    Uncompress a compressed file at `location` and return its content as a byte
    string. Raise Exceptions on errors.
    """
    with gzip.GzipFile(location, 'rb') as compressed:
        content = compressed.read()
    return content



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import os
from unittest.case import skipIf

import commoncode.testcase
from commoncode import fileutils
from commoncode import ignore
from commoncode.system import on_mac
from commoncode.system import on_windows


class IgnoreTest(commoncode.testcase.FileBasedTesting):
    test_data_dir = os.path.join(os.path.dirname(__file__), 'data')

    @skipIf(on_mac, 'Return different result on Mac for reasons to investigate')
    def test_is_ignored_default_ignores_eclipse1(self):
        test_dir = self.extract_test_tar('ignore/excludes/eclipse.tgz')
        test_base = os.path.join(test_dir, 'eclipse')

        test = os.path.join(test_base, '.settings')
        assert ignore.is_ignored(test, ignore.default_ignores, {})

    def test_is_ignored_default_ignores_eclipse2(self):
        test_dir = self.extract_test_tar('ignore/excludes/eclipse.tgz')
        test_base = os.path.join(test_dir, 'eclipse')

        test = os.path.join(test_base, '.settings/somefile')
        assert ignore.is_ignored(test, ignore.default_ignores, {})

    def test_is_ignored_default_ignores_eclipse3(self):
        test_dir = self.extract_test_tar('ignore/excludes/eclipse.tgz')
        test_base = os.path.join(test_dir, 'eclipse')

        test = os.path.join(test_base, '.project')
        assert ignore.is_ignored(test, ignore.default_ignores, {})

    def test_is_ignored_default_ignores_eclipse4(self):
        test_dir = self.extract_test_tar('ignore/excludes/eclipse.tgz')
        test_base = os.path.join(test_dir, 'eclipse')

        test = os.path.join(test_base, '.pydevproject')
        assert ignore.is_ignored(test, ignore.default_ignores, {})

    def test_is_ignored_default_ignores_mac1(self):
        test_dir = self.extract_test_tar('ignore/excludes/mac.tgz')
        test_base = os.path.join(test_dir, 'mac')

        test = os.path.join(test_base, '__MACOSX')
        assert ignore.is_ignored(test, ignore.default_ignores, {})

    def test_is_ignored_default_ignores_mac2(self):
        test_dir = self.extract_test_tar('ignore/excludes/mac.tgz')
        test_base = os.path.join(test_dir, 'mac')

        test = os.path.join(test_base, '__MACOSX/comp_match/smallrepo/._jetty_1.0_index.csv')
        assert ignore.is_ignored(test, ignore.default_ignores, {})

    def test_is_ignored_default_ignores_mac3(self):
        test_dir = self.extract_test_tar('ignore/excludes/mac.tgz')
        test_base = os.path.join(test_dir, 'mac')

        test = os.path.join(test_base, '.DS_Store')
        assert ignore.is_ignored(test, ignore.default_ignores, {})

    def test_is_ignored_default_ignores_mac4(self):
        test_dir = self.extract_test_tar('ignore/excludes/mac.tgz')
        test_base = os.path.join(test_dir, 'mac')

        test = os.path.join(test_base, '.DS_Store/a')
        assert ignore.is_ignored(test, ignore.default_ignores, {})

    @skipIf(on_mac, 'Return different result on Mac for reasons to investigate')
    def test_is_ignored_default_ignores_mac5(self):
        test_dir = self.extract_test_tar('ignore/excludes/mac.tgz')
        test_base = os.path.join(test_dir, 'mac')

        test = os.path.join(test_base, '._.DS_Store')
        # this is really weird as a behavior
        # 'Default ignore: MacOSX artifact'
        assert ignore.is_ignored(test, ignore.default_ignores, {})

    @skipIf(on_mac, 'Return different result on Mac for reasons to investigate')
    def test_is_ignored_default_ignores_msft(self):
        test_dir = self.extract_test_tar('ignore/excludes/msft-vs.tgz')
        test = os.path.join(test_dir, 'msft-vs/tst.sluo')
        # 'Default ignore: Microsoft VS project artifact' ??
        assert ignore.is_ignored(test, ignore.default_ignores, {})

    @skipIf(on_mac, 'Return different result on Mac for reasons to investigate')
    def test_is_ignored_skip_vcs_files_and_dirs(self):
        test_dir = self.extract_test_tar('ignore/vcs.tgz')
        result = []
        for top, dirs, files in os.walk(test_dir, topdown=True):
            not_ignored = []
            for d in dirs:
                p = os.path.join(top, d)
                ign = ignore.is_ignored(p, ignore.default_ignores, {})
                tp = fileutils.as_posixpath(p.replace(test_dir, ''))
                result.append((tp, ign,))
                if not ign:
                    not_ignored.append(d)

            # skip ignored things
            dirs[:] = not_ignored

            for f in files:
                p = os.path.join(top, f)
                ign = ignore.is_ignored(p, ignore.default_ignores, {})
                tp = fileutils.as_posixpath(p.replace(test_dir, ''))
                result.append((tp, ign,))

        expected = [
            ('/vcs', False),
            ('/vcs/.bzr', True),
            ('/vcs/.git', True),
            ('/vcs/.hg', True),
            ('/vcs/.repo', True),
            ('/vcs/.svn', True),
            ('/vcs/CVS', True),
            ('/vcs/_darcs', True),
            ('/vcs/_MTN', True),
            ('/vcs/.bzrignore', True),
            ('/vcs/.cvsignore', True),
            ('/vcs/.gitignore', True),
            ('/vcs/.hgignore', True),
            ('/vcs/.svnignore', True),
            ('/vcs/vssver.scc', True),
        ]
        assert sorted(result) == sorted(expected)

    def test_fileset_is_included_with_default_ignore_does_not_skip_one_char_names(self):
        # use fileset directly to work on strings not locations
        from commoncode import fileset
        tests = [c for c in 'HFS+ Private Data'] + 'HFS+ Private Data'.split()
        result = [(t,
            fileset.is_included(t, excludes=ignore.default_ignores, includes={}))
            for t in tests]
        expected = [
            ('H', True),
            ('F', True),
            ('S', True),
            ('+', True),
            (' ', False),
            ('P', True),
            ('r', True),
            ('i', True),
            ('v', True),
            ('a', True),
            ('t', True),
            ('e', True),
            (' ', False),
            ('D', True),
            ('a', True),
            ('t', True),
            ('a', True),
            ('HFS+', True),
            ('Private', True),
            ('Data', True)
        ]

        assert result == expected

    @skipIf(on_mac or on_windows, 'We are only testing on posix for now')
    def test_is_ignored_path_string_skip_special(self):
        test_path = '/test/path'
        assert ignore.is_ignored(test_path, {'asdf': 'skip'}, {}, skip_special=True)
        assert not ignore.is_ignored(test_path, {'asdf': 'skip'}, {}, skip_special=False)

    @skipIf(on_mac or on_windows, 'We are only testing on posix for now')
    def test_is_ignored_special_files_skip_special(self):
        test_fifo = self.get_temp_file()
        os.mkfifo(test_fifo)
        assert ignore.is_ignored(test_fifo, {'asdf': 'skip'}, {}, skip_special=True)
        assert not ignore.is_ignored(test_fifo, {'asdf': 'skip'}, {}, skip_special=False)

        test_symlink = self.get_temp_file()
        test_file_location = self.get_test_loc('ignore/vcs.tgz')
        os.symlink(test_file_location, test_symlink)
        assert ignore.is_ignored(test_symlink, {'asdf': 'skip'}, {}, skip_special=True)
        assert not ignore.is_ignored(test_symlink, {'asdf': 'skip'}, {}, skip_special=False)

    @skipIf(on_mac or on_windows, 'We are only testing on posix for now')
    def test_is_ignored_real_location_skip_special(self):
        test_file_location = self.get_test_loc('ignore/vcs.tgz')
        assert not ignore.is_ignored(test_file_location, {'asdf': 'skip'}, {}, skip_special=True)
        assert not ignore.is_ignored(test_file_location, {'asdf': 'skip'}, {}, skip_special=False)

        assert not ignore.is_ignored(test_file_location, {'asdf': 'skip'}, {}, skip_special=True)
        assert not ignore.is_ignored(test_file_location, {'asdf': 'skip'}, {}, skip_special=False)



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

"""
URN: Uniform Resource Name for DejaCode
=======================================

In the DejaCode platform, a URN is a globally unique and universal key to
reference data across tools and data sources. URNs are used throughout the
platform, such as in ScanCode and related tools as an object key.

URNs are both readable by humans and machines.

The URN syntax is a Internet standard specified in:
 * RFC2141 http://tools.ietf.org/html/rfc2141
 * RFC2396 http://tools.ietf.org/html/rfc2396

A DejaCode URN follows the URN syntax specification and is case sensitive. It
does support UTF-8 characters when these are URL-encoded (using the quote+
encoding).


Syntax and Examples
-------------------

The generic syntax of a DejaCode URN is::
    urn:<namespace>:<object_type>:<fields>
where:
 * <namespace> is always "dje"
 * <object_type> is a DejaCode object type. Not all objects are supported.
   Current support includes owner, license and component.
 * <fields> is one or more object-specific field each field separated by a
   colon.

The syntax for an Owner is::
    urn:dje:owner:<owner_name>
where owner_name is the name of the owner.
Example::
    urn:dje:owner:Apache+Software+Foundation

The syntax for a License is::
    urn:dje:license:<license_key>
where license_key is the key of the license.
Example::
    urn:dje:license:apache-2.0

The syntax for a Component is::
    urn:dje:component:<component_name>:<component_version>
where:
 * component_name is the name of the component
 * component_version is the version of the component. If no version is
 defined, use a trailing colon representing an empty value/undefined version.

Examples:
 * with version::
     urn:dje:component:dropbear:1.0
 * without version::
     urn:dje:component:dropbear:

The product object type syntax is the same as the component syntax.
"""

from urllib.parse import quote_plus
from urllib.parse import unquote_plus


class URNValidationError(Exception):
    """The URN format is not valid."""


# Describes the URN schema for each object type
URN_SCHEMAS = {
    'license': {
        'object': 'license',
        'fields': ['key'],
    },
    'owner': {
        'object': 'owner',
        'fields': ['name'],
    },
    'component': {
        'object': 'component',
        'fields': ['name', 'version'],
    },
    'product': {
        'object': 'product',
        'fields': ['name', 'version'],
    },
}


def encode(object_type, **fields):
    """
    Return an encoded URN based given an object_type string and a dictionary
    of the object-specific fields. All field values must be provided even if
    empty. In this case use an empty string.

    This is a string and local operation only: the URN is NOT resolved and
    therefore validity of the data for the URN as a whole and for each URN
    segment is not checked.
    """

    # case is not significant for the object type
    object_type = object_type.strip().lower()
    urn_prefix = 'urn:dje:{0}:'.format(quote_plus(object_type))

    object_fields = URN_SCHEMAS[object_type]['fields']
    # leading and trailing white spaces are not significant
    # each URN part is encoded individually BEFORE assembling the URN
    encoded_fields = [quote_plus(fields[f].strip()) for f in object_fields]
    encoded_fields = ':'.join(encoded_fields)
    return urn_prefix + encoded_fields


def decode(urn):
    """
    Decode a URN and return the object_type and a mapping of field/values.
    Raise URNValidationError on errors.
    """
    segments = [unquote_plus(p) for p in urn.split(':')]

    if not segments[0] == ('urn'):
        raise URNValidationError("Invalid URN prefix. Expected 'urn'.")

    if not segments[1] == ('dje'):
        raise URNValidationError("Invalid URN namespace. Expected 'dje'.")

    # object type is always lowercase
    object_type = segments[2].lower()
    if object_type not in URN_SCHEMAS:
        raise URNValidationError('Unsupported URN object type.')

    fields = segments[3:]
    schema_fields = URN_SCHEMAS[object_type]['fields']
    if len(schema_fields) != len(fields):
        raise URNValidationError('Invalid number of fields in URN.')
    decoded_fields = dict(zip(schema_fields, fields))

    return object_type, decoded_fields



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import os

import saneyaml

from commoncode.testcase import FileBasedTesting

TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


class TestSaneyaml(FileBasedTesting):
    test_data_dir = TEST_DATA_DIR

    def test_load_with_and_without_tags(self):
        test_file_with_tag = self.get_test_loc('saneyaml/metadata1')
        test_file_without_tag = self.get_test_loc('saneyaml/metadata1.notag')
        with_tags = saneyaml.load(open(test_file_with_tag, 'rb').read())
        without_tags = saneyaml.load(open(test_file_without_tag, 'rb').read())
        assert without_tags == with_tags



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from base64 import standard_b64decode as stddecode
from base64 import urlsafe_b64encode as b64encode

"""
Numbers to bytes or strings and URLs coder/decoders.
"""

c2i = lambda c: c
i2c = lambda i: bytes([i])


def num_to_bin(num):
    """
    Convert a `num` integer or long to a binary string byte-ordered such that
    the least significant bytes are at the beginning of the string (aka. big
    endian).
    """
    # Zero is not encoded but returned as an empty value
    if num == 0:
        return b'\x00'

    return num.to_bytes((num.bit_length() + 7) // 8, 'big')


def bin_to_num(binstr):
    """
    Convert a big endian byte-ordered binary string to an integer or long.
    """
    return int.from_bytes(binstr, byteorder='big', signed=False)


def urlsafe_b64encode(s):
    """
    Encode a binary string to a url safe base64 encoding.
    """
    return b64encode(s)


def urlsafe_b64decode(b64):
    """
    Decode a url safe base64-encoded string.
    Note that we use stddecode to work around a bug in the standard library.
    """
    b = b64.replace(b'-', b'+').replace(b'_', b'/')
    return stddecode(b)


def urlsafe_b64encode_int(num):
    """
    Encode a number (int or long) in url safe base64.
    """
    return b64encode(num_to_bin(num))



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import json
import os
from os.path import dirname
from os.path import exists
from os.path import join

from commoncode.fileutils import parent_directory
from commoncode.resource import Codebase
from commoncode.resource import depth_walk
from commoncode.resource import get_path
from commoncode.resource import VirtualCodebase

from commoncode.testcase import FileBasedTesting
from commoncode.testcase import check_against_expected_json_file


class TestCodebase(FileBasedTesting):
    test_data_dir = join(dirname(__file__), 'data')

    def test_walk_defaults(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        results = list(codebase.walk())
        expected = [
            ('codebase', False),
            ('abc', True),
            ('et131x.h', True),
            ('dir', False),
            ('that', True),
            ('this', True),
            ('other dir', False),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_Codebase_do_not_ignore_by_default_older_sccs_and_rcs_dirs(self):
        # See https://github.com/nexB/scancode-toolkit/issues/1422
        from commoncode.fileutils import create_dir

        test_codebase = self.get_temp_dir()
        create_dir(join(test_codebase, 'sccs', 'a'))
        create_dir(join(test_codebase, 'rcs', 'b'))
        codebase = Codebase(test_codebase)
        results = list(codebase.walk(topdown=True, skip_root=True))
        expected = ['rcs', 'b', 'sccs', 'a']
        assert [r.name for r in results] == expected

    def test_walk_topdown(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        results = list(codebase.walk(topdown=True))
        expected = [
            ('codebase', False),
            ('abc', True),
            ('et131x.h', True),
            ('dir', False),
            ('that', True),
            ('this', True),
            ('other dir', False),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_bottomup(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        results = list(codebase.walk(topdown=False))
        expected = [
            ('abc', True),
            ('et131x.h', True),
            ('that', True),
            ('this', True),
            ('dir', False),
            ('file', True),
            ('other dir', False),
            ('codebase', False),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_skip_root_basic(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        results = list(codebase.walk(skip_root=True))
        expected = [
            ('abc', True),
            ('et131x.h', True),
            ('dir', False),
            ('that', True),
            ('this', True),
            ('other dir', False),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_filtered_with_filtered_root(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        codebase.root.is_filtered = True
        codebase.save_resource(codebase.root)
        results = list(codebase.walk_filtered())
        expected = [
            ('abc', True),
            ('et131x.h', True),
            ('dir', False),
            ('that', True),
            ('this', True),
            ('other dir', False),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_filtered_with_all_filtered(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        for res in codebase.walk():
            res.is_filtered = True
            codebase.save_resource(res)

        results = list(codebase.walk_filtered())
        expected = []
        assert [(r.name, r.is_file) for r in results] == expected

    def test_compute_counts_filtered_None(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        results = codebase.compute_counts(skip_filtered=True)
        expected = (5, 3, 0)
        assert results == expected

    def test_compute_counts_filtered_None_with_size(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        for res in codebase.walk():
            if res.is_file:
                res.size = 10
                codebase.save_resource(res)

        results = codebase.compute_counts(skip_filtered=True)
        expected = (5, 3, 50)
        assert results == expected

    def test_compute_counts_filtered_None_with_cache(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        results = codebase.compute_counts(skip_filtered=True)
        expected = (5, 3, 0)
        assert results == expected

    def test_compute_counts_filtered_all(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        for res in codebase.walk():
            res.is_filtered = True
            codebase.save_resource(res)
        results = codebase.compute_counts(skip_filtered=True)
        expected = (0, 0, 0)
        assert results == expected

    def test_compute_counts_filtered_all_with_cache(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        for res in codebase.walk():
            res.is_filtered = True
            codebase.save_resource(res)
        results = codebase.compute_counts(skip_filtered=True)
        expected = (0, 0, 0)
        assert results == expected

    def test_compute_counts_filtered_files(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        for res in codebase.walk():
            if res.is_file:
                res.is_filtered = True
                codebase.save_resource(res)
        results = codebase.compute_counts(skip_filtered=True)
        expected = (0, 3, 0)
        assert results == expected

    def test_compute_counts_filtered_dirs(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        for res in codebase.walk():
            if not res.is_file:
                res.is_filtered = True
                codebase.save_resource(res)
        results = codebase.compute_counts(skip_filtered=True)
        expected = (5, 0, 0)
        assert results == expected

    def test_walk_filtered_dirs(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        for res in codebase.walk():
            if not res.is_file:
                res.is_filtered = True
                codebase.save_resource(res)

        results = list(codebase.walk_filtered(topdown=True))
        expected = [
            ('abc', True),
            ('et131x.h', True),
            ('that', True),
            ('this', True),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_filtered_skip_root(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        codebase.root.is_filtered = True
        codebase.save_resource(codebase.root)
        results = list(codebase.walk_filtered(skip_root=True))
        expected = [
            ('abc', True),
            ('et131x.h', True),
            ('dir', False),
            ('that', True),
            ('this', True),
            ('other dir', False),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_filtered_all_skip_root(self):
        test_codebase = self.get_test_loc('resource/codebase')
        codebase = Codebase(test_codebase)
        for res in codebase.walk():
            res.is_filtered = True
            codebase.save_resource(res)
        results = list(codebase.walk_filtered(skip_root=True))
        expected = []
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_skip_root_single_file(self):
        test_codebase = self.get_test_loc('resource/codebase/et131x.h')
        codebase = Codebase(test_codebase)
        results = list(codebase.walk(skip_root=True))
        expected = [('et131x.h', True)]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_filtered_with_skip_root_and_single_file_not_filtered(self):
        test_codebase = self.get_test_loc('resource/codebase/et131x.h')
        codebase = Codebase(test_codebase)
        results = list(codebase.walk_filtered(skip_root=True))
        expected = [('et131x.h', True)]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_filtered__with_skip_root_and_filtered_single_file(self):
        test_codebase = self.get_test_loc('resource/codebase/et131x.h')
        codebase = Codebase(test_codebase)
        codebase.root.is_filtered = True
        codebase.save_resource(codebase.root)
        results = list(codebase.walk_filtered(skip_root=True))
        expected = []
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_skip_root_single_file_with_children(self):
        test_codebase = self.get_test_loc('resource/codebase/et131x.h')
        codebase = Codebase(test_codebase, strip_root=True)

        c1 = codebase._get_or_create_resource('some child', parent=codebase.root, is_file=True)
        _c2 = codebase._get_or_create_resource('some child2', parent=c1, is_file=False)
        results = list(codebase.walk(skip_root=True))
        expected = [('some child', True), ('some child2', False)]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_filtered_with_skip_root_and_single_file_with_children(self):
        test_codebase = self.get_test_loc('resource/codebase/et131x.h')
        codebase = Codebase(test_codebase, strip_root=True)

        c1 = codebase._get_or_create_resource('some child', parent=codebase.root, is_file=True)
        c2 = codebase._get_or_create_resource('some child2', parent=c1, is_file=False)
        c2.is_filtered = True
        codebase.save_resource(c2)

        results = list(codebase.walk_filtered(skip_root=True))
        expected = [('some child', True)]
        assert [(r.name, r.is_file) for r in results] == expected

        c1.is_filtered = True
        codebase.save_resource(c1)
        results = list(codebase.walk_filtered(skip_root=True))
        expected = []
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_skip_root_single_dir(self):
        test_codebase = self.get_temp_dir('walk')
        codebase = Codebase(test_codebase, strip_root=True)

        results = list(codebase.walk(skip_root=True))
        expected = [('walk', False)]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_walk_skipped_directories_should_not_be_yielded(self):
        # Resources that we continue past should not be added to the result list
        test_codebase = self.get_test_loc('resource/skip_directories_during_walk')
        cdbs = Codebase(test_codebase)

        def _ignored(resource, codebase):
            return resource.is_dir and resource.name == 'skip-this-directory'

        result = [
            res.name
            for res in cdbs.walk(
                topdown=True,
                ignored=_ignored,
            )
        ]

        expected = ['skip_directories_during_walk', 'this-should-be-returned']
        assert result == expected

    def test__create_resource_can_add_child_to_file(self):
        test_codebase = self.get_test_loc('resource/codebase/et131x.h')
        codebase = Codebase(test_codebase)
        codebase._get_or_create_resource('some child', codebase.root, is_file=True)
        results = list(codebase.walk())
        expected = [('et131x.h', True), ('some child', True)]
        assert [(r.name, r.is_file) for r in results] == expected

    def test__create_resource_can_add_child_to_dir(self):
        test_codebase = self.get_temp_dir('resource')
        codebase = Codebase(test_codebase)
        codebase._get_or_create_resource('some child', codebase.root, is_file=False)
        results = list(codebase.walk())
        expected = [('resource', False), ('some child', False)]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_get_resource_for_single_resource_codebase(self):
        test_codebase = self.get_temp_dir('resource')
        codebase = Codebase(test_codebase)
        assert not (codebase.root is codebase.get_resource('resource'))
        assert codebase.get_resource('resource') == codebase.root

    def test_get_resource_for_multiple_resource_codebase(self):
        test_codebase = self.get_temp_dir('resource')
        for name in ('a', 'b', 'c'):
            with open(os.path.join(test_codebase, name), 'w') as o:
                o.write('\n')

        codebase = Codebase(test_codebase)
        assert codebase.get_resource('resource/a').path == 'resource/a'
        assert codebase.get_resource('/resource/c').path == 'resource/c'
        assert codebase.get_resource('resource/dsasda/../b/').path == 'resource/b'

    def test_get_path(self):
        test_dir = self.get_test_loc('resource/samples')
        locations = []
        for top, dirs, files in os.walk(test_dir):
            for x in dirs:
                locations.append(os.path.join(top, x))
            for x in files:
                locations.append(os.path.join(top, x))

        expected_default = [
            'samples/JGroups',
            'samples/zlib',
            'samples/arch',
            'samples/README',
            'samples/screenshot.png',
            'samples/JGroups/src',
            'samples/JGroups/licenses',
            'samples/JGroups/LICENSE',
            'samples/JGroups/EULA',
            'samples/JGroups/src/GuardedBy.java',
            'samples/JGroups/src/ImmutableReference.java',
            'samples/JGroups/src/RouterStub.java',
            'samples/JGroups/src/S3_PING.java',
            'samples/JGroups/src/FixedMembershipToken.java',
            'samples/JGroups/src/RouterStubManager.java',
            'samples/JGroups/src/RATE_LIMITER.java',
            'samples/JGroups/licenses/cpl-1.0.txt',
            'samples/JGroups/licenses/bouncycastle.txt',
            'samples/JGroups/licenses/lgpl.txt',
            'samples/JGroups/licenses/apache-2.0.txt',
            'samples/JGroups/licenses/apache-1.1.txt',
            'samples/zlib/dotzlib',
            'samples/zlib/iostream2',
            'samples/zlib/infback9',
            'samples/zlib/gcc_gvmat64',
            'samples/zlib/ada',
            'samples/zlib/deflate.h',
            'samples/zlib/zutil.c',
            'samples/zlib/zlib.h',
            'samples/zlib/deflate.c',
            'samples/zlib/zutil.h',
            'samples/zlib/adler32.c',
            'samples/zlib/dotzlib/AssemblyInfo.cs',
            'samples/zlib/dotzlib/LICENSE_1_0.txt',
            'samples/zlib/dotzlib/readme.txt',
            'samples/zlib/dotzlib/ChecksumImpl.cs',
            'samples/zlib/iostream2/zstream_test.cpp',
            'samples/zlib/iostream2/zstream.h',
            'samples/zlib/infback9/infback9.c',
            'samples/zlib/infback9/infback9.h',
            'samples/zlib/gcc_gvmat64/gvmat64.S',
            'samples/zlib/ada/zlib.ads',
            'samples/arch/zlib.tar.gz',
        ]

        default = sorted(get_path(test_dir, loc) for loc in locations)
        assert default == sorted(expected_default)

        expected_strip_root = [
            'JGroups',
            'zlib',
            'arch',
            'README',
            'screenshot.png',
            'JGroups/src',
            'JGroups/licenses',
            'JGroups/LICENSE',
            'JGroups/EULA',
            'JGroups/src/GuardedBy.java',
            'JGroups/src/ImmutableReference.java',
            'JGroups/src/RouterStub.java',
            'JGroups/src/S3_PING.java',
            'JGroups/src/FixedMembershipToken.java',
            'JGroups/src/RouterStubManager.java',
            'JGroups/src/RATE_LIMITER.java',
            'JGroups/licenses/cpl-1.0.txt',
            'JGroups/licenses/bouncycastle.txt',
            'JGroups/licenses/lgpl.txt',
            'JGroups/licenses/apache-2.0.txt',
            'JGroups/licenses/apache-1.1.txt',
            'zlib/dotzlib',
            'zlib/iostream2',
            'zlib/infback9',
            'zlib/gcc_gvmat64',
            'zlib/ada',
            'zlib/deflate.h',
            'zlib/zutil.c',
            'zlib/zlib.h',
            'zlib/deflate.c',
            'zlib/zutil.h',
            'zlib/adler32.c',
            'zlib/dotzlib/AssemblyInfo.cs',
            'zlib/dotzlib/LICENSE_1_0.txt',
            'zlib/dotzlib/readme.txt',
            'zlib/dotzlib/ChecksumImpl.cs',
            'zlib/iostream2/zstream_test.cpp',
            'zlib/iostream2/zstream.h',
            'zlib/infback9/infback9.c',
            'zlib/infback9/infback9.h',
            'zlib/gcc_gvmat64/gvmat64.S',
            'zlib/ada/zlib.ads',
            'arch/zlib.tar.gz',
        ]

        skipped = sorted(get_path(test_dir, loc, strip_root=True) for loc in locations)
        assert skipped == sorted(expected_strip_root)

        expected_full_ends = sorted(expected_default)
        full = sorted(get_path(test_dir, loc, full_root=True) for loc in locations)
        for full_loc, ending in zip(full, expected_full_ends):
            assert full_loc.endswith((ending))

        full_skipped = sorted(
            get_path(test_dir, loc, full_root=True, strip_root=True) for loc in locations
        )
        assert full_skipped == full

    def test_compute_counts_when_using_disk_cache(self):
        test_codebase = self.get_test_loc('resource/samples')
        codebase = Codebase(test_codebase, strip_root=True, max_in_memory=-1)
        files_count, dirs_count, size_count = codebase.compute_counts()
        assert 33 == files_count
        assert 11 == dirs_count
        assert 0 == size_count

    def test_lowest_common_parent_1(self):
        test_codebase = self.get_test_loc('resource/lcp/test1')
        codebase = Codebase(test_codebase)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'test1'
        assert lcp.name == 'test1'

    def test_lowest_common_parent_strip(self):
        test_codebase = self.get_test_loc('resource/lcp/test1')
        codebase = Codebase(test_codebase, strip_root=True)
        lcp = codebase.lowest_common_parent()
        assert lcp.name == 'test1'
        assert lcp.path == ''

    def test_lowest_common_parent_full(self):
        test_codebase = self.get_test_loc('resource/lcp/test1')
        codebase = Codebase(test_codebase, full_root=True)
        lcp = codebase.lowest_common_parent()
        assert lcp.name == 'test1'

    def test_lowest_common_parent_2(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/zlib')
        codebase = Codebase(test_codebase)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'zlib'
        assert lcp.name == 'zlib'

    def test_lowest_common_parent_3(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/simple')
        codebase = Codebase(test_codebase)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'simple'
        assert lcp.name == 'simple'

    def test_lowest_common_parent_deep(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/simple/org')
        codebase = Codebase(test_codebase)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'org/jvnet/glassfish/comms/sipagent'
        assert lcp.name == 'sipagent'

    def test_lowest_common_parent_solo_file(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/screenshot.png')
        codebase = Codebase(test_codebase)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'screenshot.png'
        assert lcp.name == 'screenshot.png'

    def test_lowest_common_parent_solo_file_strip(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/screenshot.png')
        codebase = Codebase(test_codebase, strip_root=True)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'screenshot.png'
        assert lcp.name == 'screenshot.png'

    def test_lowest_common_parent_solo_file_full(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/screenshot.png')
        codebase = Codebase(test_codebase, full_root=True)
        lcp = codebase.lowest_common_parent()
        assert lcp.name == 'screenshot.png'

    def test_distance(self):
        test_dir = self.get_test_loc('resource/dist')
        codebase = Codebase(test_dir)
        assert codebase.root.distance(test_dir) == 0

        res = codebase.get_resource('dist/JGroups')
        assert res.name == 'JGroups'
        assert res.distance(codebase) == 1

        res = codebase.get_resource('dist/simple/META-INF/MANIFEST.MF')
        assert res.name == 'MANIFEST.MF'
        assert res.distance(codebase) == 3

    def test_skip_files_and_subdirs_of_ignored_dirs(self):
        test_dir = self.get_test_loc('resource/ignore')
        codebase = Codebase(test_dir)
        # The `cvs` directory should not be visited
        expected = ['ignore', 'ignore/file1']
        result = [r.path for r in codebase.walk(topdown=True)]
        self.assertEqual(expected, result)

    def test_depth_negative_fails(self):
        test_codebase = self.get_test_loc('resource/deeply_nested')
        with self.assertRaises(Exception):
            next(depth_walk(test_codebase, -1))

    def test_depth_walk_with_depth_0(self):
        test_codebase = self.get_test_loc('resource/deeply_nested')
        results_zero = list(depth_walk(test_codebase, 0))
        results_neg = list(depth_walk(test_codebase, float("inf")))
        result_zero_dirs = [i for j in results_zero for i in j[1]]
        result_zero_files = [i for j in results_zero for i in j[2]]
        result_neg_dirs = [i for j in results_neg for i in j[1]]
        result_neg_files = [i for j in results_neg for i in j[2]]
        self.assertEqual(result_neg_dirs, result_zero_dirs)
        self.assertEqual(result_neg_files, result_zero_files)

    def test_depth_walk_with_depth_1(self):
        test_codebase = self.get_test_loc('resource/deeply_nested')
        results = list(depth_walk(test_codebase, 1))
        result_dirs = [i for j in results for i in j[1]].sort()
        result_files = [i for j in results for i in j[2]].sort()
        expected_files = ['level1_file1', 'level1_file2'].sort()
        expected_dirs = ['level1_dir1', 'level1_dir2'].sort()
        self.assertEqual(result_dirs, expected_dirs)
        self.assertEqual(result_files, expected_files)

    def test_depth_walk_with_depth_2(self):
        test_codebase = self.get_test_loc('resource/deeply_nested')
        results = list(depth_walk(test_codebase, 2))
        result_dirs = [i for j in results for i in j[1]].sort()
        result_files = [i for j in results for i in j[2]].sort()
        expected_files = [
            'level1_file1',
            'level1_file2',
            'level2_file2',
            'level2_file1',
            'level2_file3',
            'level2_file4',
            'level2_file5',
        ].sort()
        expected_dirs = ['level1_dir1', 'level1_dir2', 'level2_dir1', 'level2_dir3'].sort()
        self.assertEqual(result_dirs, expected_dirs)
        self.assertEqual(result_files, expected_files)

    def test_depth_walk_with_depth_3(self):
        test_codebase = self.get_test_loc('resource/deeply_nested')
        results = list(depth_walk(test_codebase, 3))
        result_dirs = [i for j in results for i in j[1]].sort()
        result_files = [i for j in results for i in j[2]].sort()
        expected_files = [
            'level1_file1',
            'level1_file2',
            'level2_file2',
            'level2_file1',
            'level3_file2',
            'level3_file1',
            'level2_file3',
            'level2_file4',
            'level2_file5',
            'level3_file4',
            'level3_file3',
        ].sort()
        expected_dirs = [
            'level1_dir1',
            'level1_dir2',
            'level2_dir1',
            'level3_dir1',
            'level2_dir3',
        ].sort()
        self.assertEqual(result_dirs, expected_dirs)
        self.assertEqual(result_files, expected_files)

    def test_specify_depth_1(self):
        test_codebase = self.get_test_loc('resource/deeply_nested')
        codebase = Codebase(test_codebase, max_depth=1)
        results = list(codebase.walk())
        expected = [
            ('deeply_nested', False),
            ('level1_dir1', False),
            ('level1_dir2', False),
            ('level1_file1', True),
            ('level1_file2', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_specify_depth_2(self):
        test_codebase = self.get_test_loc('resource/deeply_nested')
        codebase = Codebase(test_codebase, max_depth=2)
        results = list(codebase.walk())

        expected = [
            ('deeply_nested', False),
            ('level1_file1', True),
            ('level1_file2', True),
            ('level1_dir1', False),
            ('level2_dir1', False),
            ('level2_file1', True),
            ('level2_file2', True),
            ('level1_dir2', False),
            ('level2_dir3', False),
            ('level2_file3', True),
            ('level2_file4', True),
            ('level2_file5', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_specify_depth_3(self):
        test_codebase = self.get_test_loc('resource/deeply_nested')
        codebase = Codebase(test_codebase, max_depth=3)
        results = list(codebase.walk())

        expected = [
            ('deeply_nested', False),
            ('level1_file1', True),
            ('level1_file2', True),
            ('level1_dir1', False),
            ('level2_file1', True),
            ('level2_file2', True),
            ('level2_dir1', False),
            ('level3_dir1', False),
            ('level3_file1', True),
            ('level3_file2', True),
            ('level1_dir2', False),
            ('level2_file3', True),
            ('level2_file4', True),
            ('level2_file5', True),
            ('level2_dir3', False),
            ('level3_file3', True),
            ('level3_file4', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected


class TestCodebaseWithPath(FileBasedTesting):
    test_data_dir = join(dirname(__file__), 'data')

    def test_Codebase_with_paths_works(self):
        test_codebase = self.get_test_loc('resource/with_path/codebase')
        paths = ['codebase/other dir/file']
        codebase = Codebase(location=test_codebase, paths=paths)
        assert not codebase.errors
        results = [r.to_dict() for r in codebase.walk()]
        print(r.path for r in codebase.walk())
        expected_file = self.get_test_loc(
            'resource/with_path/codebase-expected.json',
            must_exist=False,
        )
        check_against_expected_json_file(results, expected_file, regen=False)

    def test_VirtualCodebase_with_paths_works(self):
        test_codebase = self.get_test_loc('resource/with_path/virtual-codebase.json')
        paths = ['codebase/other dir/file']
        codebase = VirtualCodebase(location=test_codebase, paths=paths)
        assert not codebase.errors
        results = [r.to_dict() for r in codebase.walk()]
        expected_file = self.get_test_loc(
            'resource/with_path/virtual-codebase-expected.json',
            must_exist=False,
        )
        check_against_expected_json_file(results, expected_file, regen=False)


class TestCodebaseCache(FileBasedTesting):
    test_data_dir = join(dirname(__file__), 'data')

    def test_codebase_cache_default(self):
        test_codebase = self.get_test_loc('resource/cache2')
        codebase = Codebase(test_codebase)

        assert codebase.temp_dir
        assert codebase.cache_dir

        root = codebase.root

        cp = codebase._get_resource_cache_location(root.path, create_dirs=True)
        assert not exists(cp)
        assert exists(parent_directory(cp))

        child = codebase._get_or_create_resource(name='child', parent=root, is_file=True)
        child.size = 12
        codebase.save_resource(child)
        child_2 = codebase.get_resource(path=child.path)
        assert child_2 == child

    def test_codebase_cache_all_in_memory(self):
        test_codebase = self.get_test_loc('resource/cache2')
        codebase = Codebase(test_codebase, max_in_memory=0)
        for path, res in codebase.resources_by_path.items():
            if res is Codebase.CACHED_RESOURCE:
                res = codebase.get_resource(path)
            if res.is_root:
                assert codebase.get_resource(path) == codebase.root == res
                assert codebase._exists_in_memory(path)
                assert not codebase._exists_on_disk(path)
            else:
                assert codebase._exists_in_memory(path)
                assert not codebase._exists_on_disk(path)

        assert (
            len(list(codebase.walk()))
            == len(codebase.resources_by_path)
            == codebase.resources_count
        )

    def test_codebase_cache_all_on_disk(self):
        test_codebase = self.get_test_loc('resource/cache2')
        codebase = Codebase(test_codebase, max_in_memory=-1)
        for path, res in codebase.resources_by_path.items():
            if res is Codebase.CACHED_RESOURCE:
                res = codebase.get_resource(path)
            if res.is_root:
                assert codebase.get_resource(path) == codebase.root == res
                assert codebase._exists_in_memory(path)
                assert not codebase._exists_on_disk(path)
            else:
                assert not codebase._exists_in_memory(path)
                assert codebase._exists_on_disk(path)

        assert (
            len(list(codebase.walk()))
            == len(codebase.resources_by_path)
            == codebase.resources_count
        )

    def test_codebase_cache_mixed_two_in_memory(self):
        test_codebase = self.get_test_loc('resource/cache2')
        codebase = Codebase(test_codebase, max_in_memory=2)
        counter = 0
        for path, res in codebase.resources_by_path.items():
            if res is Codebase.CACHED_RESOURCE:
                res = codebase.get_resource(path)

            if res.is_root:
                assert codebase.get_resource(path) == codebase.root == res
                assert codebase._exists_in_memory(path)
                assert not codebase._exists_on_disk(path)
                counter += 1
            elif counter < 2:
                assert codebase._exists_in_memory(path)
                assert not codebase._exists_on_disk(path)
                counter += 1
            else:
                assert not codebase._exists_in_memory(path)
                assert codebase._exists_on_disk(path)

        assert (
            len(list(codebase.walk()))
            == len(codebase.resources_by_path)
            == codebase.resources_count
        )


class TestVirtualCodebase(FileBasedTesting):
    test_data_dir = join(dirname(__file__), 'data')

    def test_virtual_codebase_walk_defaults(self):
        test_file = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        codebase = VirtualCodebase(location=test_file)
        results = list(codebase.walk())
        expected = [
            ('codebase', False),
            ('abc', True),
            ('et131x.h', True),
            ('dir', False),
            ('that', True),
            ('this', True),
            ('other dir', False),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_walk_topdown(self):
        test_file = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        codebase = VirtualCodebase(location=test_file)
        results = list(codebase.walk(topdown=True))
        expected = [
            ('codebase', False),
            ('abc', True),
            ('et131x.h', True),
            ('dir', False),
            ('that', True),
            ('this', True),
            ('other dir', False),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_walk_bottomup(self):
        test_file = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        codebase = VirtualCodebase(location=test_file)
        results = list(codebase.walk(topdown=False))
        expected = [
            ('abc', True),
            ('et131x.h', True),
            ('that', True),
            ('this', True),
            ('dir', False),
            ('file', True),
            ('other dir', False),
            ('codebase', False),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_walk_skip_root_basic(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        results = list(virtual_codebase.walk(skip_root=True))
        expected = [
            ('abc', True),
            ('et131x.h', True),
            ('dir', False),
            ('that', True),
            ('this', True),
            ('other dir', False),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_get_path_with_strip_root_and_walk_with_skip_root(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/stripped-and-skipped-root.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        results = [r.get_path(strip_root=True) for r in virtual_codebase.walk(skip_root=True)]
        expected = ['README', 'screenshot.png']
        assert expected == results

    def test_virtual_codebase_walk_filtered_with_filtered_root(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        virtual_codebase.root.is_filtered = True
        virtual_codebase.save_resource(virtual_codebase.root)

        results = list(virtual_codebase.walk_filtered())
        expected = [
            ('abc', True),
            ('et131x.h', True),
            ('dir', False),
            ('that', True),
            ('this', True),
            ('other dir', False),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_walk_filtered_with_all_filtered(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        for res in virtual_codebase.walk():
            res.is_filtered = True
            virtual_codebase.save_resource(res)

        results = list(virtual_codebase.walk_filtered())
        expected = []
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_compute_counts_filtered_None(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        results = virtual_codebase.compute_counts(skip_filtered=True)
        expected = (5, 3, 2228)
        assert results == expected

    def test_virtual_codebase_compute_counts_filtered_None_with_size(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        for res in virtual_codebase.walk():
            if res.is_file:
                res.size = 10
                virtual_codebase.save_resource(res)

        results = virtual_codebase.compute_counts(skip_filtered=True)
        expected = (5, 3, 50)
        assert results == expected

    def test_virtual_codebase_compute_counts_filtered_None_with_cache(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        results = virtual_codebase.compute_counts(skip_filtered=True)
        expected = (5, 3, 2228)
        assert results == expected

    def test_virtual_codebase_compute_counts_filtered_all(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        for res in virtual_codebase.walk():
            res.is_filtered = True
            virtual_codebase.save_resource(res)
        results = virtual_codebase.compute_counts(skip_filtered=True)
        expected = (0, 0, 0)
        assert results == expected

    def test_virtual_codebase_compute_counts_filtered_all_with_cache(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        for res in virtual_codebase.walk():
            res.is_filtered = True
            virtual_codebase.save_resource(res)
        results = virtual_codebase.compute_counts(skip_filtered=True)
        expected = (0, 0, 0)
        assert results == expected

    def test_virtual_codebase_compute_counts_filtered_files(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        for res in virtual_codebase.walk():
            if res.is_file:
                res.is_filtered = True
                virtual_codebase.save_resource(res)
        results = virtual_codebase.compute_counts(skip_filtered=True)
        expected = (0, 3, 0)
        assert results == expected

    def test_virtual_codebase_compute_counts_filtered_dirs(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        for res in virtual_codebase.walk():
            if not res.is_file:
                res.is_filtered = True
                virtual_codebase.save_resource(res)
        results = virtual_codebase.compute_counts(skip_filtered=True)
        expected = (5, 0, 2228)
        assert results == expected

    def test_virtual_codebase_walk_filtered_dirs(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        for res in virtual_codebase.walk():
            if not res.is_file:
                res.is_filtered = True
                virtual_codebase.save_resource(res)
        results = list(virtual_codebase.walk_filtered(topdown=True))
        expected = [
            ('abc', True),
            ('et131x.h', True),
            ('that', True),
            ('this', True),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_walk_filtered_skip_root(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        virtual_codebase.root.is_filtered = True
        virtual_codebase.save_resource(virtual_codebase.root)
        results = list(virtual_codebase.walk_filtered(skip_root=True))
        expected = [
            ('abc', True),
            ('et131x.h', True),
            ('dir', False),
            ('that', True),
            ('this', True),
            ('other dir', False),
            ('file', True),
        ]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_walk_filtered_all_skip_root(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/virtual_codebase.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        for res in virtual_codebase.walk():
            res.is_filtered = True
            virtual_codebase.save_resource(res)

        results = list(virtual_codebase.walk_filtered(skip_root=True))
        expected = []
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_walk_skip_root_single_file(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/et131x.h.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        results = list(virtual_codebase.walk(skip_root=True))
        expected = [('et131x.h', True)]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_walk_filtered_with_skip_root_and_single_file_not_filtered(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/et131x.h.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        results = list(virtual_codebase.walk_filtered(skip_root=True))
        expected = [('et131x.h', True)]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_walk_filtered__with_skip_root_and_filtered_single_file(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/et131x.h.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        virtual_codebase.root.is_filtered = True
        virtual_codebase.save_resource(virtual_codebase.root)
        results = list(virtual_codebase.walk_filtered(skip_root=True))
        expected = []
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_walk_skip_root_single_file_with_children(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/et131x.h.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        c1 = virtual_codebase._get_or_create_resource(
            'some child', parent=virtual_codebase.root, is_file=True
        )
        _c2 = virtual_codebase._get_or_create_resource('some child2', parent=c1, is_file=False)
        results = list(virtual_codebase.walk(skip_root=True))
        expected = [('some child', True), ('some child2', False)]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_walk_filtered_with_skip_root_and_single_file_with_children(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/et131x.h.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        c1 = virtual_codebase._get_or_create_resource(
            'some child', parent=virtual_codebase.root, is_file=True
        )
        c2 = virtual_codebase._get_or_create_resource('some child2', parent=c1, is_file=False)
        c2.is_filtered = True
        virtual_codebase.save_resource(c2)

        results = list(virtual_codebase.walk_filtered(skip_root=True))
        expected = [('some child', True)]
        assert [(r.name, r.is_file) for r in results] == expected

        c1.is_filtered = True
        virtual_codebase.save_resource(c1)
        results = list(virtual_codebase.walk_filtered(skip_root=True))
        expected = []
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase__create_resource_can_add_child_to_file(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/et131x.h.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        virtual_codebase._get_or_create_resource('some child', virtual_codebase.root, is_file=True)
        results = list(virtual_codebase.walk())
        expected = [('et131x.h', True), ('some child', True)]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase__create_resource_can_add_child_to_dir(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/resource.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        virtual_codebase._get_or_create_resource('some child', virtual_codebase.root, is_file=False)
        results = list(virtual_codebase.walk())
        expected = [('resource', False), ('some child', False)]
        assert [(r.name, r.is_file) for r in results] == expected

    def test_virtual_codebase_get_resource(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/resource.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        assert not (virtual_codebase.root is virtual_codebase.get_resource('resource'))
        assert virtual_codebase.get_resource('resource') == virtual_codebase.root

    def test_virtual_codebase_can_process_minimal_resources_without_info(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/noinfo.json')
        codebase = VirtualCodebase(location=scan_data)
        expected = [
            {
                'path': 'NOTICE',
                'type': 'file',
                'copyrights': [
                    {
                        'statements': ['Copyright (c) 2017 nexB Inc. and others.'],
                        'holders': ['nexB Inc. and others.'],
                        'authors': [],
                        'start_line': 4,
                        'end_line': 4,
                    }
                ],
                'scan_errors': [],
            }
        ]
        assert [r.to_dict() for r in codebase.walk()] == expected

    def test_virtual_codebase_can_process_minimal_resources_with_only_path(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/only-path.json')
        codebase = VirtualCodebase(location=scan_data)
        expected = [
            {'path': 'samples', 'type': 'directory', 'scan_errors': []},
            {'path': 'samples/NOTICE', 'type': 'file', 'scan_errors': []},
        ]
        assert [r.to_dict() for r in codebase.walk()] == expected

    def test_VirtualCodebase_account_fingerprint_attribute(self):
        test_file = self.get_test_loc("resource/virtual_codebase/fingerprint_attribute.json")
        codebase = VirtualCodebase(test_file)
        resources_fingerprint = [resource.fingerprint for resource in codebase.walk()]
        assert "e30cf09443e7878dfed3288886e97542" in resources_fingerprint
        assert None in resources_fingerprint
        assert codebase.get_resource('apache_to_all_notable_lic_new') == codebase.root
        assert resources_fingerprint.count(None) == 2

    def test_VirtualCodebase_works_with_mapping_backed_codebase(self):
        test_file = self.get_test_loc("resource/virtual_codebase/license-scan.json")
        codebase = VirtualCodebase(test_file)
        resource = codebase.get_resource('scan-ref/license-notice.txt')
        assert resource
        assert len(resource.license_expressions) == 1


class TestCodebaseLowestCommonParent(FileBasedTesting):
    test_data_dir = join(dirname(__file__), 'data')

    def test_lowest_common_parent_on_virtual_codebase(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/lcp.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        lcp = virtual_codebase.lowest_common_parent()
        assert lcp.path == 'lcp/test1'
        assert lcp.name == 'test1'

    def test_virtual_codebase_has_default_for_plugin_attributes(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/only-path.json')
        VirtualCodebase(location=scan_data)

    def test_lowest_common_parent_1(self):
        test_codebase = self.get_test_loc('resource/lcp/test1')
        codebase = Codebase(test_codebase)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'test1'
        assert lcp.name == 'test1'

    def test_lowest_common_parent_strip(self):
        test_codebase = self.get_test_loc('resource/lcp/test1')
        codebase = Codebase(test_codebase, strip_root=True)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == ''
        assert lcp.name == 'test1'

    def test_lowest_common_parent_full(self):
        test_codebase = self.get_test_loc('resource/lcp/test1')
        codebase = Codebase(test_codebase, full_root=True)
        lcp = codebase.lowest_common_parent()
        assert lcp.name == 'test1'

    def test_lowest_common_parent_2(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/zlib')
        codebase = Codebase(test_codebase)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'zlib'
        assert lcp.name == 'zlib'

    def test_lowest_common_parent_3(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/simple')
        codebase = Codebase(test_codebase)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'simple'
        assert lcp.name == 'simple'

    def test_lowest_common_parent_deep(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/simple/org')
        codebase = Codebase(test_codebase)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'org/jvnet/glassfish/comms/sipagent'
        assert lcp.name == 'sipagent'

    def test_lowest_common_parent_solo_file(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/screenshot.png')
        codebase = Codebase(test_codebase)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'screenshot.png'
        assert lcp.name == 'screenshot.png'

    def test_lowest_common_parent_solo_file_strip(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/screenshot.png')
        codebase = Codebase(test_codebase, strip_root=True)
        lcp = codebase.lowest_common_parent()
        assert lcp.path == 'screenshot.png'
        assert lcp.name == 'screenshot.png'

    def test_lowest_common_parent_solo_file_full(self):
        test_codebase = self.get_test_loc('resource/lcp/test1/screenshot.png')
        codebase = Codebase(test_codebase, full_root=True)
        lcp = codebase.lowest_common_parent()
        assert lcp.name == 'screenshot.png'


class TestVirtualCodebaseCache(FileBasedTesting):
    test_data_dir = join(dirname(__file__), 'data')

    def test_virtual_codebase_cache_default(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/codebase-for-cache-tests.json')
        virtual_codebase = VirtualCodebase(location=scan_data)
        assert virtual_codebase.temp_dir
        assert virtual_codebase.cache_dir
        virtual_codebase.cache_dir
        root = virtual_codebase.root

        cp = virtual_codebase._get_resource_cache_location(root.path, create_dirs=False)
        assert not exists(cp)

        cp = virtual_codebase._get_resource_cache_location(root.path, create_dirs=True)
        assert not exists(cp)
        assert exists(parent_directory(cp))

        child = virtual_codebase._get_or_create_resource('child', root, is_file=True)
        child.size = 12
        virtual_codebase.save_resource(child)
        child_2 = virtual_codebase.get_resource(child.path)
        assert child_2 == child

    def test_virtual_codebase_cache_all_in_memory(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/codebase-for-cache-tests.json')
        virtual_codebase = VirtualCodebase(location=scan_data, max_in_memory=0)
        for path, res in virtual_codebase.resources_by_path.items():
            assert res != Codebase.CACHED_RESOURCE
            if res.is_root:
                assert virtual_codebase.get_resource(path).to_dict(with_info=True) == virtual_codebase.root.to_dict(with_info=True)
                assert virtual_codebase._exists_in_memory(path)
                assert not virtual_codebase._exists_on_disk(path)
            else:
                assert virtual_codebase._exists_in_memory(path)
                assert not virtual_codebase._exists_on_disk(path)

        assert (
            len(list(virtual_codebase.walk()))
            == len(virtual_codebase.resources_by_path)
            == virtual_codebase.resources_count
        )

    def test_virtual_codebase_cache_all_on_disk(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/codebase-for-cache-tests.json')
        virtual_codebase = VirtualCodebase(location=scan_data, max_in_memory=-1)
        for path, res in virtual_codebase.resources_by_path.items():

            if res != Codebase.CACHED_RESOURCE:
                assert res.is_root
            else:
                res = virtual_codebase.get_resource(path)

            if res.is_root:
                assert virtual_codebase.get_resource(path) == virtual_codebase.root
                assert virtual_codebase._exists_in_memory(path)
                assert not virtual_codebase._exists_on_disk(path)
            else:
                assert not virtual_codebase._exists_in_memory(path)
                assert virtual_codebase._exists_on_disk(path)

        assert (
            len(list(virtual_codebase.walk()))
            == len(virtual_codebase.resources_by_path)
            == virtual_codebase.resources_count
        )

    def test_virtual_codebase_cache_mixed_two_in_memory(self):
        scan_data = self.get_test_loc('resource/virtual_codebase/codebase-for-cache-tests.json')
        virtual_codebase = VirtualCodebase(location=scan_data, max_in_memory=2)
        counter = 0

        for path, res in virtual_codebase.resources_by_path.items():
            if res is Codebase.CACHED_RESOURCE:
                res = virtual_codebase.get_resource(path)

            if res.is_root:
                assert virtual_codebase.get_resource(path).to_dict() == virtual_codebase.root.to_dict()
                assert virtual_codebase._exists_in_memory(path)
                assert not virtual_codebase._exists_on_disk(path)
                counter += 1

            elif counter < 2:
                assert virtual_codebase._exists_in_memory(path)
                assert not virtual_codebase._exists_on_disk(path)
                counter += 1

            else:
                assert not virtual_codebase._exists_in_memory(path)
                assert virtual_codebase._exists_on_disk(path)

        assert (
            len(list(virtual_codebase.walk()))
            == len(virtual_codebase.resources_by_path)
            == virtual_codebase.resources_count
        )


class TestVirtualCodebaseCreation(FileBasedTesting):
    test_data_dir = join(dirname(__file__), 'data')

    def test_VirtualCodebase_can_be_created_from_json_file(self):
        test_file = self.get_test_loc('resource/virtual_codebase/from_file.json')
        codebase = VirtualCodebase(test_file)
        results = sorted(r.name for r in codebase.walk())
        expected = ['bar.svg', 'han']
        assert results == expected

    def test_VirtualCodebase_can_be_created_from_json_string(self):
        test_scan = '''
            {
              "scancode_notice": "Generated with ScanCode and provided on an ....",
              "scancode_version": "2.9.7.post137.2e29fe3.dirty.20181120225811",
              "scancode_options": {
                "input": "han/",
                "--json-pp": "-"
              },
              "scan_start": "2018-11-23T123252.191917",
              "files_count": 1,
              "files": [
                {
                  "path": "han",
                  "type": "directory",
                  "scan_errors": []
                },
                {
                  "path": "han/bar.svg",
                  "type": "file",
                  "scan_errors": []
                }
              ]
            }
            '''
        codebase = VirtualCodebase(test_scan)
        results = sorted(r.name for r in codebase.walk())
        expected = ['bar.svg', 'han']
        assert results == expected

    def test_VirtualCodebase_can_be_created_from_dict(self):
        test_scan = {
            "scancode_notice": "Generated with ScanCode and provided on an ....",
            "scancode_version": "2.9.7.post137.2e29fe3.dirty.20181120225811",
            "scancode_options": {"input": "han/", "--json-pp": "-"},
            "scan_start": "2018-11-23T123252.191917",
            "files_count": 1,
            "files": [
                {"path": "han", "type": "directory", "scan_errors": []},
                {"path": "han/bar.svg", "type": "file", "scan_errors": []},
            ],
        }
        codebase = VirtualCodebase(test_scan)

        results = sorted(r.name for r in codebase.walk())
        expected = ['bar.svg', 'han']
        assert results == expected

    def test_VirtualCodebase_create_from_scan_with_no_root_and_missing_parents(self):
        test_file = self.get_test_loc('resource/virtual_codebase/samples-only-findings.json')
        result_file = self.get_test_loc(
            'resource/virtual_codebase/samples-only-findings-expected.json'
        )
        codebase = VirtualCodebase(test_file)
        expected_scan = json.load(open(result_file))
        results = sorted(r.path for r in codebase.walk())
        expected = sorted(r.get('path') for r in expected_scan['files'])
        assert results == expected

    def test_VirtualCodebase_check_that_already_existing_parent_is_updated_properly(self):
        test_file = self.get_test_loc('resource/virtual_codebase/root-is-not-first-resource.json')
        codebase = VirtualCodebase(test_file)
        results = sorted((r.to_dict() for r in codebase.walk()), key=lambda x: tuple(x.items()))
        expected = [
            {'path': 'samples', 'type': 'directory', 'summary': ['asd'], 'scan_errors': []},
            {'path': 'samples/NOTICE', 'type': 'file', 'summary': [], 'scan_errors': []},
        ]
        assert results == expected

    def test_VirtualCodebase_create_from_multiple_scans(self):
        test_file_1 = self.get_test_loc('resource/virtual_codebase/combine-1.json')
        test_file_2 = self.get_test_loc('resource/virtual_codebase/combine-2.json')
        vinput = (test_file_1, test_file_2)
        codebase = VirtualCodebase(vinput)
        results = [r.to_dict(with_info=False) for r in codebase.walk()]
        expected_file = self.get_test_loc(
            'resource/virtual_codebase/combine-expected.json',
            must_exist=False,
        )
        check_against_expected_json_file(results, expected_file, regen=False)

    def test_VirtualCodebase_create_from_multiple_scans_shared_directory_names(self):
        test_file_1 = self.get_test_loc(
            'resource/virtual_codebase/combine-shared-directory-name-1.json'
        )
        test_file_2 = self.get_test_loc(
            'resource/virtual_codebase/combine-shared-directory-name-2.json'
        )
        vinput = (test_file_1, test_file_2)
        codebase = VirtualCodebase(location=vinput)

        results = [r.to_dict(with_info=False) for r in codebase.walk()]
        expected_file = self.get_test_loc(
            'resource/virtual_codebase/combine-shared-directory-name-expected.json',
            must_exist=False,
        )
        check_against_expected_json_file(results, expected_file, regen=False)

    def test_VirtualCodebase_compute_counts_with_full_root_info_one(self):
        test_file = self.get_test_loc('resource/virtual_codebase/full-root-info-one.json')
        codebase = VirtualCodebase(test_file)
        resource = [r for r in codebase.walk() if r.is_file][0]
        assert resource.path == 'home/foobar/scancode-toolkit/samples/README'
        files_count, dirs_count, size_count = codebase.compute_counts()
        assert files_count == 1
        assert dirs_count == 0
        assert size_count == 236

    def test_VirtualCodebase_with_full_root_info_one(self):
        test_file = self.get_test_loc('resource/virtual_codebase/full-root-info-one.json')
        codebase = VirtualCodebase(test_file)
        results = [r.to_dict(with_info=True) for r in codebase.walk()]
        expected_file = self.get_test_loc('resource/virtual_codebase/full-root-info-one-expected.json', must_exist=False)
        check_against_expected_json_file(results, expected_file, regen=False)

    def test_VirtualCodebase_with_full_root_info_many(self):
        test_file = self.get_test_loc('resource/virtual_codebase/full-root-info-many.json')
        codebase = VirtualCodebase(test_file)
        results = [r.to_dict(with_info=True) for r in codebase.walk()]
        expected_file = self.get_test_loc('resource/virtual_codebase/full-root-info-many-expected.json', must_exist=False)
        check_against_expected_json_file(results, expected_file, regen=False)

    def test_VirtualCodebase_can_compute_counts_with_null(self):
        # was failing with
        # size_count += child.size
        # TypeError: unsupported operand type(s) for +=: 'int' and 'NoneType'
        test_file = self.get_test_loc('resource/virtual_codebase/node-16-slim.json')
        codebase = VirtualCodebase(test_file)
        codebase.compute_counts()

    def test_VirtualCodebase_can_be_created_with_single_path(self):
        test_file = self.get_test_loc('resource/virtual_codebase/docker-hello-world.json')
        VirtualCodebase(test_file)

    def test_VirtualCodebase_can_be_created_without_RecursionError(self):
        # was failing with RecursionError: maximum recursion depth exceeded
        test_file = self.get_test_loc('resource/virtual_codebase/zephyr-binary.json')
        VirtualCodebase(test_file)


class TestResource(FileBasedTesting):
    test_data_dir = join(dirname(__file__), 'data')

    def test_Resource_extracted_to_extracted_from(self):
        test_file = self.get_test_loc('resource/resource/test-extracted-from-to.json')
        codebase = VirtualCodebase(location=test_file)
        results = []
        for r in codebase.walk(topdown=True):
            extracted_to = r.extracted_to(codebase)
            extracted_from = r.extracted_from(codebase)

            extracted_to_path = extracted_to and extracted_to.path
            extracted_from_path = extracted_from and extracted_from.path
            results.append((r.path, extracted_to_path, extracted_from_path))

        expected = [
            (
                'test',
                None,
                None,
            ),
            (
                'test/c',
                None,
                None,
            ),
            (
                'test/foo.tar.gz',
                'test/foo.tar.gz-extract',
                None,
            ),
            (
                'test/foo.tar.gz-extract',
                None,
                'test/foo.tar.gz',
            ),
            (
                'test/foo.tar.gz-extract/foo',
                None,
                'test/foo.tar.gz',
            ),
            (
                'test/foo.tar.gz-extract/foo/a',
                None,
                'test/foo.tar.gz',
            ),
            (
                'test/foo.tar.gz-extract/foo/bar.tar.gz',
                'test/foo.tar.gz-extract/foo/bar.tar.gz-extract',
                'test/foo.tar.gz',
            ),
            (
                'test/foo.tar.gz-extract/foo/bar.tar.gz-extract',
                None,
                'test/foo.tar.gz-extract/foo/bar.tar.gz',
            ),
            (
                'test/foo.tar.gz-extract/foo/bar.tar.gz-extract/bar',
                None,
                'test/foo.tar.gz-extract/foo/bar.tar.gz',
            ),
            (
                'test/foo.tar.gz-extract/foo/bar.tar.gz-extract/bar/b',
                None,
                'test/foo.tar.gz-extract/foo/bar.tar.gz',
            ),
        ]
        assert results == expected

    def test_virtualcode_Resource_can_walk(self):
        test_file = self.get_test_loc('resource/resource/test-extracted-from-to.json')
        codebase = VirtualCodebase(location=test_file)
        results = [r.path for r in codebase.walk(topdown=True)]

        expected = [
            'test',
            'test/c',
            'test/foo.tar.gz',
            'test/foo.tar.gz-extract',
            'test/foo.tar.gz-extract/foo',
            'test/foo.tar.gz-extract/foo/a',
            'test/foo.tar.gz-extract/foo/bar.tar.gz',
            'test/foo.tar.gz-extract/foo/bar.tar.gz-extract',
            'test/foo.tar.gz-extract/foo/bar.tar.gz-extract/bar',
            'test/foo.tar.gz-extract/foo/bar.tar.gz-extract/bar/b',
        ]

        assert results == expected



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import getpass
import os
import sys


def os_arch():
    """
    Return a tuple for the current the OS and architecture.
    """
    if sys.maxsize > 2 ** 32:
        arch = '64'
    else:
        arch = '32'

    sys_platform = str(sys.platform).lower()
    if sys_platform.startswith('linux'):
        os = 'linux'
    elif 'win32' in sys_platform:
        os = 'win'
    elif 'darwin' in sys_platform:
        os = 'mac'
    elif 'freebsd' in sys_platform:
        os = 'freebsd'
    else:
        raise Exception('Unsupported OS/platform %r' % sys_platform)
    return os, arch


#
# OS/Arch
#
current_os, current_arch = os_arch()
on_windows = current_os == 'win'
on_windows_32 = on_windows and current_arch == '32'
on_windows_64 = on_windows and current_arch == '64'
on_mac = current_os == 'mac'
on_linux = current_os == 'linux'
on_freebsd = current_os == 'freebsd'
on_posix = not on_windows and (on_mac or on_linux or on_freebsd)

current_os_arch = '%(current_os)s-%(current_arch)s' % locals()
noarch = 'noarch'
current_os_noarch = '%(current_os)s-%(noarch)s' % locals()

del os_arch


def is_on_macos_14_or_higher():
    """
    Return True if the current OS is macOS 14 or higher.
    It uses APFS by default and has a different behavior wrt. unicode and
    filesystem encodings.
    """
    import platform
    macos_ver = platform.mac_ver()
    macos_ver = macos_ver[0]
    macos_ver = macos_ver.split('.')
    return macos_ver > ['10', '14']


on_macos_14_or_higher = is_on_macos_14_or_higher()

del is_on_macos_14_or_higher


def has_case_sensitive_fs():
    """
    Return True if the current FS is case sensitive.

    Windows is not case sensitive, and while older macOS HPFS+ were POSIX and
    case sensitive by default, newer macOS use APFS which is no longer case
    sensitive by default.

    From https://developer.apple.com/library/archive/documentation/FileManagement/Conceptual/APFS_Guide/FAQ/FAQ.html
        How does Apple File System handle filenames?
        APFS accepts only valid UTF-8 encoded filenames for creation, and preserves
        both case and normalization of the filename on disk in all variants. APFS,
        like HFS+, is case-sensitive on iOS and is available in case-sensitive and
        case-insensitive variants on macOS, with case-insensitive being the default.
    """
    return not os.path.exists(__file__.upper())


is_case_sensitive_fs = has_case_sensitive_fs()

#
# Shared library file extensions
#
if on_windows:
    lib_ext = '.dll'
if on_mac:
    lib_ext = '.dylib'
if on_linux or on_freebsd:
    lib_ext = '.so'

#
# Python versions
#
_sys_v0 = sys.version_info[0]
py2 = _sys_v0 == 2
py3 = _sys_v0 == 3

_sys_v1 = sys.version_info[1]
py36 = py3 and _sys_v1 == 6
py37 = py3 and _sys_v1 == 7
py38 = py3 and _sys_v1 == 8
py39 = py3 and _sys_v1 == 9
py310 = py3 and _sys_v1 == 10

# Do not let Windows error pop up messages with default SetErrorMode
# See http://msdn.microsoft.com/en-us/library/ms680621(VS100).aspx
#
# SEM_FAILCRITICALERRORS:
# The system does not display the critical-error-handler message box.
# Instead, the system sends the error to the calling process.
#
# SEM_NOGPFAULTERRORBOX:
# The system does not display the Windows Error Reporting dialog.
if on_windows:
    import ctypes
    # 3 is SEM_FAILCRITICALERRORS | SEM_NOGPFAULTERRORBOX
    ctypes.windll.kernel32.SetErrorMode(3)  # @UndefinedVariable
